from __future__ import annotations


def function_with_py310_annotations(
    self,  # noqa: ANN001
    x: bool | None,
    y: int | str | float,  # noqa: PYI041
    z: str | None = None,
) -> str:
    """
    Method docstring.

    :param x: foo
    :param y: bar
    :param z: baz
    """



from __future__ import annotations

import re
import sys
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest
from conftest import normalize_sphinx_text

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, AsyncIterator, Generator, Iterator
    from io import StringIO

    from sphinx.testing.util import SphinxTestApp


class GenClass:
    """A class with generators."""

    def gen_method(  # noqa: PLR6301
        self,
        arg1: int | None = None,  # noqa: ARG002
    ) -> Generator[tuple[str, ...], None, None]:
        """Summary.

        Args:
            arg1: argument 1

        Yields:
            strings of things
        """
        yield ("test",)

    def iter_method(self) -> Iterator[int]:  # noqa: PLR6301
        """Summary.

        Yields:
            integers
        """
        yield 1

    async def async_gen_method(self) -> AsyncGenerator[str, None]:  # noqa: PLR6301
        """Summary.

        Yields:
            strings
        """
        yield "test"

    async def async_iter_method(self) -> AsyncIterator[int]:  # noqa: PLR6301
        """Summary.

        Yields:
            integers
        """
        yield 1

    def no_yields_method(self) -> Generator[int, None, None]:  # noqa: PLR6301
        """Summary without yields section."""
        yield 1


def _extract_section(text: str, method_name: str) -> str:
    pattern = rf"({method_name}\(\).*?)(?=(?:async\s+)?\w+\(\)|$)"
    if match := re.search(pattern, text, re.DOTALL):
        return match.group(1)
    msg = f"Method {method_name} not found in output"
    raise ValueError(msg)


def _build_genclass(app: SphinxTestApp, monkeypatch: pytest.MonkeyPatch) -> str:
    (Path(app.srcdir) / "index.rst").write_text(
        dedent("""\
        Test
        ====

        .. autoclass:: mod.GenClass
           :members:
    """)
    )
    monkeypatch.setitem(sys.modules, "mod", sys.modules[__name__])
    app.build()
    return normalize_sphinx_text((Path(app.srcdir) / "_build/text/index.txt").read_text())


@pytest.mark.sphinx("text", testroot="integration")
def test_generator_no_duplicate_return_type(
    app: SphinxTestApp,
    status: StringIO,
    warning: StringIO,  # noqa: ARG001
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    text = _build_genclass(app, monkeypatch)
    assert "build succeeded" in status.getvalue()

    for method in ("gen_method", "iter_method", "async_gen_method", "async_iter_method"):
        section = _extract_section(text, method)
        assert "Yields" in section, f"{method} missing Yields"
        assert "Return type" not in section, f"{method} has unexpected Return type"


@pytest.mark.sphinx("text", testroot="integration")
def test_generator_without_yields_keeps_rtype(
    app: SphinxTestApp,
    status: StringIO,
    warning: StringIO,  # noqa: ARG001
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    text = _build_genclass(app, monkeypatch)
    assert "build succeeded" in status.getvalue()

    section = _extract_section(text, "no_yields_method")
    assert "Return type" in section



from __future__ import annotations

from unittest.mock import MagicMock, patch

from sphinx_autodoc_typehints.attributes_patch import (
    OKAY_TO_PATCH,
    TYPE_IS_RST_LABEL,
    patch_attribute_handling,
    patched_parse_annotation,
)


def test_okay_to_patch_is_true() -> None:
    assert OKAY_TO_PATCH


def test_patched_parse_annotation_without_label() -> None:
    settings = MagicMock()
    env = MagicMock()
    with patch("sphinx_autodoc_typehints.attributes_patch._parse_annotation") as mock_parse:
        mock_parse.return_value = "parsed"
        result = patched_parse_annotation(settings, "int", env)
    assert result == "parsed"
    mock_parse.assert_called_once_with("int", env)


def test_patched_parse_annotation_dispatches_on_label() -> None:
    settings = MagicMock()
    env = MagicMock()
    with patch("sphinx_autodoc_typehints.attributes_patch.rst_to_docutils", return_value=["nodes"]) as mock_rst:
        result = patched_parse_annotation(settings, f"{TYPE_IS_RST_LABEL}**bold**", env)
    mock_rst.assert_called_once_with(settings, "**bold**")
    assert result == ["nodes"]


def test_patch_attribute_handling_when_not_okay() -> None:
    app = MagicMock()
    with patch("sphinx_autodoc_typehints.attributes_patch.OKAY_TO_PATCH", False):
        patch_attribute_handling(app)


def test_rst_to_docutils_returns_parsed_nodes() -> None:
    """Lines 34-36: rst_to_docutils parses RST and returns inner nodes."""
    from docutils.frontend import get_default_settings  # noqa: PLC0415
    from sphinx.parsers import RSTParser  # noqa: PLC0415

    from sphinx_autodoc_typehints.attributes_patch import rst_to_docutils  # noqa: PLC0415

    settings = get_default_settings(RSTParser)  # type: ignore[arg-type]
    settings.env = MagicMock()
    result = rst_to_docutils(settings, "**bold text**")
    assert len(result) > 0
    assert any("bold text" in node.astext() for node in result)



from __future__ import annotations

import pathlib
import sys

sys.path.insert(0, str(pathlib.Path(__file__).parent))

master_doc = "index"
extensions = [
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",
]



from __future__ import annotations

import sys
from pathlib import Path
from textwrap import dedent, indent
from typing import TYPE_CHECKING, Any, TypeVar, Union
from unittest.mock import MagicMock

import pytest
from conftest import normalize_sphinx_text
from sphinx.util.inspect import TypeAliasForwardRef

from sphinx_autodoc_typehints import format_annotation

if TYPE_CHECKING:
    from collections.abc import Callable
    from io import StringIO

    from sphinx.testing.util import SphinxTestApp

T = TypeVar("T")

UserId = Union[int, str]
RequestData = dict[str, Any]

TYPE_PREAMBLE = """\
type mod.UserId

   A user identifier that can be either an integer or a string.

type mod.RequestData

   Request data dictionary.

"""


def expected(expected: str, **options: Any) -> Callable[[T], T]:
    def dec(val: T) -> T:
        val.EXPECTED = expected  # ty: ignore[unresolved-attribute]
        val.OPTIONS = options  # ty: ignore[unresolved-attribute]
        return val

    return dec


@expected(
    TYPE_PREAMBLE
    + """\
mod.get_user(user_id)

   Get a user by ID.

   Parameters:
      **user_id** ("UserId") -- The user identifier

   Return type:
      "str"
"""
)
def get_user(user_id: UserId) -> str:
    """
    Get a user by ID.

    Args:
        user_id: The user identifier
    """
    return f"User {user_id}"


@expected(
    TYPE_PREAMBLE
    + """\
mod.process_request(data)

   Process a request.

   Parameters:
      **data** ("RequestData") -- The request data

   Return type:
      "bool"
"""
)
def process_request(data: RequestData) -> bool:  # noqa: ARG001
    """
    Process a request.

    Args:
        data: The request data
    """
    return True


configs = {"default_conf": {}}


@pytest.mark.parametrize("val", [x for x in globals().values() if hasattr(x, "EXPECTED")])
@pytest.mark.parametrize("conf_run", list(configs.keys()))
@pytest.mark.sphinx("text", testroot="issue_599")
def test_integration(
    app: SphinxTestApp,
    status: StringIO,
    warning: StringIO,
    monkeypatch: pytest.MonkeyPatch,
    val: Any,
    conf_run: str,
) -> None:
    template = """\
.. py:type:: mod.UserId

   A user identifier that can be either an integer or a string.

.. py:type:: mod.RequestData

   Request data dictionary.

.. autofunction:: mod.{}
"""

    (Path(app.srcdir) / "index.rst").write_text(template.format(val.__name__))
    app.config.__dict__.update(configs[conf_run])
    app.config.__dict__.update(val.OPTIONS)
    monkeypatch.setitem(sys.modules, "mod", sys.modules[__name__])
    app.build()
    assert "build succeeded" in status.getvalue()

    value = warning.getvalue().strip()
    assert not value or "Inline strong start-string without end-string" in value

    result = normalize_sphinx_text((Path(app.srcdir) / "_build/text/index.txt").read_text())

    expected_text = normalize_sphinx_text(val.EXPECTED)
    try:
        assert result.strip() == dedent(expected_text).strip()
    except Exception:
        indented = indent(f'"""\n{result}\n"""', " " * 4)
        print(f"@expected(\n{indented}\n)\n")  # noqa: T201
        raise


@pytest.mark.sphinx("text", testroot="issue_599")
def test_eager_annotations(
    app: SphinxTestApp,
    status: StringIO,
    warning: StringIO,
) -> None:
    """Test that non-deferred annotations (no ``from __future__ import annotations``) also resolve."""
    template = """\
.. py:type:: mod_eager.UserId

   A user identifier.

.. autofunction:: mod_eager.get_user_eager
"""
    (Path(app.srcdir) / "index.rst").write_text(template)
    app.build()
    assert "build succeeded" in status.getvalue()

    value = warning.getvalue().strip()
    assert not value or "Inline strong start-string without end-string" in value

    result = normalize_sphinx_text((Path(app.srcdir) / "_build/text/index.txt").read_text())
    assert '"UserId"' in result


def test_format_annotation_type_alias_without_env() -> None:
    """TypeAliasForwardRef falls back to plain name when no env is available."""
    config = MagicMock()
    config.typehints_formatter = None
    del config._typehints_env  # noqa: SLF001
    annotation = TypeAliasForwardRef("SomeAlias")
    assert format_annotation(annotation, config) == "SomeAlias"



from __future__ import annotations

from unittest.mock import patch

from sphinx_autodoc_typehints._resolver._util import get_obj_location


def _typed_func(x: int, y: str) -> bool: ...


class _BackfillClass:
    def method(self, x: int) -> str: ...


def test_get_obj_location_with_function() -> None:
    location = get_obj_location(_typed_func)
    assert location is not None
    assert location.endswith(".py:" + str(_typed_func.__code__.co_firstlineno))
    assert "test_util.py" in location


def test_get_obj_location_with_class() -> None:
    location = get_obj_location(_BackfillClass)
    assert location is not None
    assert "test_util.py" in location


def test_get_obj_location_non_inspectable() -> None:
    assert get_obj_location(42) is None


def test_get_obj_location_no_source_lines() -> None:
    with patch("sphinx_autodoc_typehints._resolver._util.inspect.getsourcelines", side_effect=OSError):
        location = get_obj_location(_typed_func)
    assert location is not None
    assert location.endswith(".py")
    assert ":" not in location.rsplit("/", 1)[-1]



from __future__ import annotations

import re
import sys
from functools import cmp_to_key
from pathlib import Path
from textwrap import dedent, indent
from typing import TYPE_CHECKING, Any
from unittest.mock import create_autospec, patch

import pytest
import typing_extensions
from conftest import normalize_sphinx_text
from sphinx.application import Sphinx
from sphinx.config import Config
from sphinx.ext.autodoc import Options

from sphinx_autodoc_typehints import process_docstring, process_signature

if TYPE_CHECKING:
    from io import StringIO
    from types import FunctionType

    from sphinx.testing.util import SphinxTestApp


class Slotted:
    __slots__ = ()


class HintedMethods:
    @classmethod
    def from_magic(cls) -> typing_extensions.Self: ...

    def method(self) -> typing_extensions.Self: ...


def test_process_docstring_slot_wrapper() -> None:
    lines: list[str] = []
    config = create_autospec(
        Config,
        typehints_fully_qualified=False,
        simplify_optional_unions=False,
        typehints_formatter=None,
        autodoc_mock_imports=[],
    )
    app: Sphinx = create_autospec(Sphinx, config=config)
    process_docstring(app, "class", "SlotWrapper", Slotted, None, lines)
    assert not lines


def test_process_docstring_wrapper_loop() -> None:
    """Regression test for #405: inspect.unwrap raises ValueError on wrapper loops."""

    def func(x: int) -> str: ...

    func.__wrapped__ = func  # type: ignore[attr-defined]  # circular wrapper loop

    lines: list[str] = []
    config = create_autospec(
        Config,
        typehints_fully_qualified=False,
        simplify_optional_unions=False,
        typehints_formatter=None,
        autodoc_mock_imports=[],
    )
    app: Sphinx = create_autospec(Sphinx, config=config)
    process_docstring(app, "function", "func", func, None, lines)


def test_process_signature_wrapper_loop() -> None:
    """Regression test for #405: inspect.unwrap raises ValueError on wrapper loops."""

    def func(x: int) -> str: ...

    func.__wrapped__ = func  # type: ignore[attr-defined]  # circular wrapper loop

    config = create_autospec(
        Config,
        typehints_fully_qualified=False,
        simplify_optional_unions=False,
        typehints_formatter=None,
        typehints_use_signature=False,
        typehints_use_signature_return=False,
        autodoc_type_aliases={},
    )
    app: Sphinx = create_autospec(Sphinx, config=config)
    result = process_signature(
        app,
        "function",
        "func",
        func,
        Options(),
        "",
        "",
    )
    assert result is None


def set_python_path() -> None:
    test_path = Path(__file__).parent
    if str(test_path) not in sys.path:
        sys.path.insert(0, str(test_path))


@pytest.mark.parametrize("always_document_param_types", [True, False], ids=["doc_param_type", "no_doc_param_type"])
@pytest.mark.sphinx("text", testroot="dummy")
@patch("sphinx.writers.text.MAXWIDTH", 2000)
def test_always_document_param_types(
    app: SphinxTestApp,
    status: StringIO,
    warning: StringIO,
    always_document_param_types: bool,
) -> None:
    set_python_path()

    app.config.always_document_param_types = always_document_param_types  # create flag
    app.config.autodoc_mock_imports = ["mailbox"]  # create flag

    for f in Path(app.srcdir).glob("*.rst"):
        f.unlink()
    (Path(app.srcdir) / "index.rst").write_text(
        dedent(
            """
            .. autofunction:: dummy_module.undocumented_function

            .. autoclass:: dummy_module.DataClass
                :undoc-members:
                :special-members: __init__
            """,
        ),
    )

    app.build()

    assert "build succeeded" in status.getvalue()
    assert not warning.getvalue().strip()

    format_args = {}
    for indentation_level in range(2):
        key = f"undoc_params_{indentation_level}"
        if always_document_param_types:
            format_args[key] = indent('\n\n   Parameters:\n      **x** ("int")', "   " * indentation_level)
        else:
            format_args[key] = ""

    contents = normalize_sphinx_text((Path(app.srcdir) / "_build/text/index.txt").read_text())
    expected_contents = """\
    dummy_module.undocumented_function(x)

       Hi{undoc_params_0}

       Return type:
          "str"

    class dummy_module.DataClass(x)

       Class docstring.{undoc_params_0}

       __init__(x){undoc_params_1}
    """
    expected_contents = normalize_sphinx_text(dedent(expected_contents).format(**format_args))
    assert contents == expected_contents


@pytest.mark.sphinx("text", testroot="dummy")
@patch("sphinx.writers.text.MAXWIDTH", 2000)
def test_always_document_param_types_with_defaults_braces_after(
    app: SphinxTestApp,
    status: StringIO,
    warning: StringIO,  # noqa: ARG001
) -> None:
    """Regression test for #575: IndexError when combining always_document_param_types with braces-after."""
    set_python_path()

    app.config.always_document_param_types = True
    app.config.typehints_defaults = "braces-after"

    for rst_file in Path(app.srcdir).glob("*.rst"):
        rst_file.unlink()
    index_content = """\
        .. autofunction:: dummy_module.undocumented_function_with_defaults
    """
    (Path(app.srcdir) / "index.rst").write_text(dedent(index_content))

    app.build()

    assert "build succeeded" in status.getvalue()


@pytest.mark.sphinx("text", testroot="dummy")
@patch("sphinx.writers.text.MAXWIDTH", 2000)
def test_namedtuple_new_no_warning(
    app: SphinxTestApp,
    status: StringIO,
    warning: StringIO,
) -> None:
    """Regression test for #601: NamedTuple __new__ causes 'NoneType' attribute error."""
    set_python_path()

    for rst_file in Path(app.srcdir).glob("*.rst"):
        rst_file.unlink()
    index_content = """\
        .. autoclass:: dummy_module.MyNamedTuple
            :special-members: __new__
    """
    (Path(app.srcdir) / "index.rst").write_text(dedent(index_content))

    app.build()

    assert "build succeeded" in status.getvalue()
    assert "NoneType" not in warning.getvalue()


@pytest.mark.sphinx("text", testroot="dummy")
@patch("sphinx.writers.text.MAXWIDTH", 2000)
def test_sphinx_output_future_annotations(app: SphinxTestApp, status: StringIO) -> None:
    set_python_path()

    app.config.master_doc = "future_annotations"  # create flag
    app.build()

    assert "build succeeded" in status.getvalue()

    contents = normalize_sphinx_text((Path(app.srcdir) / "_build/text/future_annotations.txt").read_text())
    expected_contents = """\
    Dummy Module
    ************

    dummy_module_future_annotations.function_with_py310_annotations(self, x, y, z=None)

       Method docstring.

       Parameters:
          * **x** ("bool" | "None") -- foo

          * **y** ("int" | "str" | "float") -- bar

          * **z** ("str" | "None") -- baz

       Return type:
          "str"
    """
    expected_contents = normalize_sphinx_text(dedent(expected_contents))
    assert contents == expected_contents


@pytest.mark.sphinx("pseudoxml", testroot="dummy")
@patch("sphinx.writers.text.MAXWIDTH", 2000)
def test_sphinx_output_default_role(app: SphinxTestApp, status: StringIO) -> None:
    set_python_path()

    app.config.master_doc = "simple_default_role"  # create flag
    app.config.default_role = "literal"
    app.build()

    assert "build succeeded" in status.getvalue()

    contents_lines = (
        (Path(app.srcdir) / "_build/pseudoxml/simple_default_role.pseudoxml").read_text(encoding="utf-8").splitlines()
    )
    list_item_idxs = [i for i, line in enumerate(contents_lines) if line.strip() == "<list_item>"]
    foo_param = dedent("\n".join(contents_lines[list_item_idxs[0] : list_item_idxs[1]]))
    expected_foo_param = """\
    <list_item>
        <paragraph>
            <literal_strong>
                x
             (
            <inline classes="sphinx_autodoc_typehints-type">
                <literal classes="xref py py-class">
                    bool
            )
             \N{EN DASH}\N{SPACE}
            <literal>
                foo
    """.rstrip()
    expected_foo_param = dedent(expected_foo_param)
    assert foo_param == expected_foo_param


@pytest.mark.parametrize(
    ("defaults_config_val", "expected"),
    [
        (None, '("int") -- bar'),
        ("comma", '("int", default: "1") -- bar'),
        ("braces", '("int" (default: "1")) -- bar'),
        ("braces-after", '("int") -- bar (default: "1")'),
        ("comma-after", Exception("needs to be one of")),
    ],
)
@pytest.mark.sphinx("text", testroot="dummy")
@patch("sphinx.writers.text.MAXWIDTH", 2000)
def test_sphinx_output_defaults(
    app: SphinxTestApp,
    status: StringIO,
    defaults_config_val: str,
    expected: str | Exception,
) -> None:
    set_python_path()

    app.config.master_doc = "simple"  # create flag
    app.config.typehints_defaults = defaults_config_val  # create flag
    if isinstance(expected, Exception):
        with pytest.raises(Exception, match=re.escape(str(expected))):
            app.build()
        return
    app.build()
    assert "build succeeded" in status.getvalue()

    contents = normalize_sphinx_text((Path(app.srcdir) / "_build/text/simple.txt").read_text())
    expected_contents = f"""\
    Simple Module
    *************

    dummy_module_simple.function(x, y=1)

       Function docstring.

       Parameters:
          * **x** ("bool") -- foo

          * **y** {expected}

       Return type:
          "str"
    """
    assert contents == normalize_sphinx_text(dedent(expected_contents))


@pytest.mark.parametrize(
    ("formatter_config_val", "expected"),
    [
        (None, ['("bool") -- foo', '("int") -- bar', '"str"']),
        (lambda ann, conf: "Test", ["(Test) -- foo", "(Test) -- bar", "Test"]),  # noqa: ARG005
        ("some string", Exception("needs to be callable or `None`")),
    ],
)
@pytest.mark.sphinx("text", testroot="dummy")
@patch("sphinx.writers.text.MAXWIDTH", 2000)
def test_sphinx_output_formatter(
    app: SphinxTestApp,
    status: StringIO,
    formatter_config_val: str,
    expected: tuple[str, ...] | Exception,
) -> None:
    set_python_path()

    app.config.master_doc = "simple"  # create flag
    app.config.typehints_formatter = formatter_config_val  # create flag
    if isinstance(expected, Exception):
        with pytest.raises(Exception, match=re.escape(str(expected))):
            app.build()
        return
    app.build()
    assert "build succeeded" in status.getvalue()

    contents = normalize_sphinx_text((Path(app.srcdir) / "_build/text/simple.txt").read_text())
    expected_contents = f"""\
    Simple Module
    *************

    dummy_module_simple.function(x, y=1)

       Function docstring.

       Parameters:
          * **x** {expected[0]}

          * **y** {expected[1]}

       Return type:
          {expected[2]}
    """
    assert contents == normalize_sphinx_text(dedent(expected_contents))


@pytest.mark.parametrize("obj", [cmp_to_key, 1])
def test_default_no_signature(obj: Any) -> None:
    config = create_autospec(
        Config,
        typehints_fully_qualified=False,
        simplify_optional_unions=False,
        typehints_formatter=None,
        autodoc_mock_imports=[],
    )
    app: Sphinx = create_autospec(Sphinx, config=config)
    lines: list[str] = []
    process_docstring(app, "what", "name", obj, None, lines)
    assert lines == []


@pytest.mark.parametrize("method", [HintedMethods.from_magic, HintedMethods().method])
def test_bound_class_method(method: FunctionType) -> None:
    config = create_autospec(
        Config,
        typehints_fully_qualified=False,
        simplify_optional_unions=False,
        typehints_document_rtype=False,
        always_document_param_types=True,
        typehints_defaults=True,
        typehints_formatter=None,
        autodoc_mock_imports=[],
    )
    app: Sphinx = create_autospec(Sphinx, config=config)
    process_docstring(app, "class", method.__qualname__, method, None, [])


@pytest.mark.sphinx("text", testroot="resolve-typing-guard")
def test_resolve_typing_guard_imports(app: SphinxTestApp, status: StringIO, warning: StringIO) -> None:
    set_python_path()
    app.config.autodoc_mock_imports = ["viktor"]  # create flag
    app.build()
    out = status.getvalue()
    assert "build succeeded" in out
    assert "Failed guarded type import" not in warning.getvalue()


@pytest.mark.sphinx("text", testroot="resolve-typing-guard-tmp")
def test_resolve_typing_guard_attrs_imports(app: SphinxTestApp, status: StringIO, warning: StringIO) -> None:
    set_python_path()
    app.build()
    assert "build succeeded" in status.getvalue()
    assert not warning.getvalue()


@pytest.mark.sphinx("text", testroot="dummy")
@patch("sphinx.writers.text.MAXWIDTH", 2000)
def test_sphinx_output_formatter_no_use_rtype(app: SphinxTestApp, status: StringIO) -> None:
    set_python_path()
    app.config.master_doc = "simple_no_use_rtype"  # create flag
    app.config.typehints_use_rtype = False
    app.build()
    assert "build succeeded" in status.getvalue()
    text_path = Path(app.srcdir) / "_build" / "text" / "simple_no_use_rtype.txt"
    text_contents = normalize_sphinx_text(text_path.read_text())
    expected_contents = """\
    Simple Module
    *************

    dummy_module_simple_no_use_rtype.function_no_returns(x, y=1)

       Function docstring.

       Parameters:
          * **x** ("bool") -- foo

          * **y** ("int") -- bar

       Return type:
          "str"

    dummy_module_simple_no_use_rtype.function_returns_with_type(x, y=1)

       Function docstring.

       Parameters:
          * **x** ("bool") -- foo

          * **y** ("int") -- bar

       Returns:
          *CustomType* -- A string

    dummy_module_simple_no_use_rtype.function_returns_with_compound_type(x, y=1)

       Function docstring.

       Parameters:
          * **x** ("bool") -- foo

          * **y** ("int") -- bar

       Returns:
          Union[str, int] -- A string or int

    dummy_module_simple_no_use_rtype.function_returns_without_type(x, y=1)

       Function docstring.

       Parameters:
          * **x** ("bool") -- foo

          * **y** ("int") -- bar

       Returns:
          "str" -- A string
    """
    assert text_contents == normalize_sphinx_text(dedent(expected_contents))


@pytest.mark.sphinx("text", testroot="dummy")
@patch("sphinx.writers.text.MAXWIDTH", 2000)
def test_sphinx_output_with_use_signature(app: SphinxTestApp, status: StringIO) -> None:
    set_python_path()
    app.config.master_doc = "simple"  # create flag
    app.config.typehints_use_signature = True
    app.build()
    assert "build succeeded" in status.getvalue()
    text_path = Path(app.srcdir) / "_build" / "text" / "simple.txt"
    text_contents = normalize_sphinx_text(text_path.read_text())
    expected_contents = """\
    Simple Module
    *************

    dummy_module_simple.function(x: bool, y: int = 1)

       Function docstring.

       Parameters:
          * **x** ("bool") -- foo

          * **y** ("int") -- bar

       Return type:
          "str"
    """
    assert text_contents == normalize_sphinx_text(dedent(expected_contents))


@pytest.mark.sphinx("text", testroot="dummy")
@patch("sphinx.writers.text.MAXWIDTH", 2000)
def test_sphinx_output_with_use_signature_return(app: SphinxTestApp, status: StringIO) -> None:
    set_python_path()
    app.config.master_doc = "simple"  # create flag
    app.config.typehints_use_signature_return = True
    app.build()
    assert "build succeeded" in status.getvalue()
    text_path = Path(app.srcdir) / "_build" / "text" / "simple.txt"
    text_contents = normalize_sphinx_text(text_path.read_text())
    expected_contents = """\
    Simple Module
    *************

    dummy_module_simple.function(x, y=1) -> str

       Function docstring.

       Parameters:
          * **x** ("bool") -- foo

          * **y** ("int") -- bar

       Return type:
          "str"
    """
    assert text_contents == normalize_sphinx_text(dedent(expected_contents))


@pytest.mark.sphinx("text", testroot="dummy")
@patch("sphinx.writers.text.MAXWIDTH", 2000)
def test_sphinx_output_with_use_signature_and_return(app: SphinxTestApp, status: StringIO) -> None:
    set_python_path()
    app.config.master_doc = "simple"  # create flag
    app.config.typehints_use_signature = True
    app.config.typehints_use_signature_return = True
    app.build()
    assert "build succeeded" in status.getvalue()
    text_path = Path(app.srcdir) / "_build" / "text" / "simple.txt"
    text_contents = normalize_sphinx_text(text_path.read_text())
    expected_contents = """\
    Simple Module
    *************

    dummy_module_simple.function(x: bool, y: int = 1) -> str

       Function docstring.

       Parameters:
          * **x** ("bool") -- foo

          * **y** ("int") -- bar

       Return type:
          "str"
    """
    assert text_contents == normalize_sphinx_text(dedent(expected_contents))


@pytest.mark.sphinx("text", testroot="dummy")
@patch("sphinx.writers.text.MAXWIDTH", 2000)
def test_default_annotation_without_typehints(app: SphinxTestApp, status: StringIO) -> None:
    set_python_path()
    app.config.master_doc = "without_complete_typehints"  # create flag
    app.config.typehints_defaults = "comma"
    app.build()
    assert "build succeeded" in status.getvalue()
    text_path = Path(app.srcdir) / "_build" / "text" / "without_complete_typehints.txt"
    text_contents = normalize_sphinx_text(text_path.read_text())
    expected_contents = """\
    Simple Module
    *************

    dummy_module_without_complete_typehints.function_with_some_defaults_and_without_typehints(x, y=None)

       Function docstring.

       Parameters:
          * **x** -- foo

          * **y** (default: "None") -- bar

    dummy_module_without_complete_typehints.function_with_some_defaults_and_some_typehints(x, y=None)

       Function docstring.

       Parameters:
          * **x** ("int") -- foo

          * **y** (default: "None") -- bar

    dummy_module_without_complete_typehints.function_with_some_defaults_and_more_typehints(x, y=None)

       Function docstring.

       Parameters:
          * **x** ("int") -- foo

          * **y** (default: "None") -- bar

       Return type:
          "str"

    dummy_module_without_complete_typehints.function_with_defaults_and_some_typehints(x=0, y=None)

       Function docstring.

       Parameters:
          * **x** ("int", default: "0") -- foo

          * **y** (default: "None") -- bar

       Return type:
          "str"

    dummy_module_without_complete_typehints.function_with_defaults_and_type_information_in_docstring(x, y=0)

       Function docstring.

       Parameters:
          * **x** ("int") -- foo

          * **y** (int, default: "0") -- bar

       Return type:
          "str"
    """
    assert text_contents == normalize_sphinx_text(dedent(expected_contents))


@pytest.mark.sphinx("text", testroot="dummy")
@patch("sphinx.writers.text.MAXWIDTH", 2000)
def test_wrong_module_path(app: SphinxTestApp, status: StringIO, warning: StringIO) -> None:
    set_python_path()

    app.config.master_doc = "wrong_module_path"  # create flag
    app.config.default_role = "literal"
    app.config.nitpicky = True
    app.config.nitpick_ignore = {("py:data", "typing.Optional")}

    def fixup_module_name(mod: str) -> str:
        if not mod.startswith("wrong_module_path"):
            return mod
        return "export_module" + mod.removeprefix("wrong_module_path")

    app.config.suppress_warnings = ["config.cache"]
    app.config.typehints_fixup_module_name = fixup_module_name
    app.build()

    assert "build succeeded" in status.getvalue()
    assert not warning.getvalue().strip()



from __future__ import annotations

import importlib.util
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Annotated

import pytest
from conftest import normalize_sphinx_text
from typing_extensions import Doc

from sphinx_autodoc_typehints import _extract_doc_description

if TYPE_CHECKING:
    from io import StringIO

    from sphinx.testing.util import SphinxTestApp

numpydoc = pytest.importorskip("numpydoc")


@pytest.mark.parametrize(
    ("annotation", "expected"),
    [
        pytest.param(Annotated[int, Doc("hello")], "hello", id="annotated-with-doc"),
        pytest.param(Annotated[int, 42], None, id="annotated-without-doc"),
        pytest.param(int, None, id="plain-type"),
        pytest.param(Annotated[int, Doc("first"), Doc("second")], "first", id="picks-first-doc"),
    ],
)
def test_extract_doc_description(annotation: type, expected: str | None) -> None:
    assert _extract_doc_description(annotation) == expected


def _load_and_build(
    app: SphinxTestApp,
    status: StringIO,
    monkeypatch: pytest.MonkeyPatch,
    testroot: str,
    func_name: str,
) -> str:
    mod_name = "mod"
    mod_path = Path(__file__).parent / "roots" / f"test-{testroot}" / f"{mod_name}.py"
    spec = importlib.util.spec_from_file_location(mod_name, mod_path)
    assert spec is not None
    assert spec.loader is not None
    module = importlib.util.module_from_spec(spec)
    sys.modules[mod_name] = module
    spec.loader.exec_module(module)
    (Path(app.srcdir) / "index.rst").write_text(f".. autofunction:: {mod_name}.{func_name}\n")
    monkeypatch.setitem(sys.modules, mod_name, module)
    app.build()
    assert "build succeeded" in status.getvalue()
    return normalize_sphinx_text((Path(app.srcdir) / "_build/text/index.txt").read_text())


@pytest.mark.parametrize(
    ("func_name", "expected", "not_expected"),
    [
        pytest.param(
            "greet",
            ["The person's name", "The greeting phrase", "The full greeting message"],
            [],
            id="all-params-and-return",
        ),
        pytest.param("partial_doc", ["The x value", "The y value"], [], id="partial-doc-with-docstring"),
        pytest.param("no_doc", ["Identity"], [], id="annotated-without-doc-metadata"),
        pytest.param("docstring_wins", ["Docstring description"], ["Doc description"], id="docstring-takes-precedence"),
    ],
)
@pytest.mark.sphinx("text", testroot="annotated-doc")
def test_sphinx_field_list(
    app: SphinxTestApp,
    status: StringIO,
    monkeypatch: pytest.MonkeyPatch,
    func_name: str,
    expected: list[str],
    not_expected: list[str],
) -> None:
    result = _load_and_build(app, status, monkeypatch, "annotated-doc", func_name)
    for text in expected:
        assert text in result
    for text in not_expected:
        assert text not in result


@pytest.mark.parametrize(
    ("func_name", "expected"),
    [
        pytest.param("transform", ["The input data", "The transformed result"], id="doc-injected"),
        pytest.param("compute", ["Placeholder", "The sum"], id="existing-params-preserved"),
    ],
)
@pytest.mark.sphinx("text", testroot="annotated-doc-numpydoc")
def test_numpydoc(
    app: SphinxTestApp,
    status: StringIO,
    monkeypatch: pytest.MonkeyPatch,
    func_name: str,
    expected: list[str],
) -> None:
    result = _load_and_build(app, status, monkeypatch, "annotated-doc-numpydoc", func_name)
    for text in expected:
        assert text in result



from __future__ import annotations

import re
import shutil
import sys
from contextlib import suppress
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from sphobjinv import Inventory

if TYPE_CHECKING:
    from _pytest.config import Config

pytest_plugins = "sphinx.testing.fixtures"
collect_ignore = ["roots"]

_SPHINX_EMPHASIS_RE = re.compile(
    r"""
    (?<!\*\*)   # not preceded by ** (bold markup)
    (?<!\*)     # not preceded by * (avoids partial match inside **)
    \*          # opening *
    ([^*\s]     # first char: not * or whitespace
    [^*]*?)     # rest: non-greedy, no * chars
    \*          # closing *
    (?!\*)      # not followed by *
    """,
    re.VERBOSE,
)


def normalize_sphinx_text(text: str) -> str:
    text = text.replace("\u2013", "--")
    return _SPHINX_EMPHASIS_RE.sub(r'"\1"', text)


@pytest.fixture(scope="session")
def inv(pytestconfig: Config) -> Inventory:
    cache_path = f"python{sys.version_info.major}.{sys.version_info.minor}/objects.inv"
    assert pytestconfig.cache is not None
    inv_dict = pytestconfig.cache.get(cache_path, None)
    if inv_dict is not None:
        return Inventory(inv_dict)  # ty: ignore[too-many-positional-arguments]

    url = f"https://docs.python.org/{sys.version_info.major}.{sys.version_info.minor}/objects.inv"
    inv = Inventory(url=url)  # ty: ignore[unknown-argument]
    pytestconfig.cache.set(cache_path, inv.json_dict())
    return inv


@pytest.fixture(autouse=True)
def _remove_sphinx_projects(sphinx_test_tempdir: Path) -> None:
    # Remove any directory which appears to be a Sphinx project from
    # the temporary directory area.
    # See https://github.com/sphinx-doc/sphinx/issues/4040
    for entry in sphinx_test_tempdir.iterdir():
        with suppress(PermissionError):
            if entry.is_dir() and Path(entry, "_build").exists():
                shutil.rmtree(str(entry))


@pytest.fixture
def rootdir() -> Path:
    return Path(str(Path(__file__).parent) or ".").absolute() / "roots"


def pytest_ignore_collect(collection_path: Path, config: Config) -> bool | None:  # noqa: ARG001
    version_re = re.compile(r"_py(\d)(\d)\.py$")
    match = version_re.search(collection_path.name)
    if match:
        version = tuple(int(x) for x in match.groups())
        if sys.version_info < version:  # ty: ignore[unsupported-operator]
            return True
    return None



from __future__ import annotations

import inspect
from typing import Any
from unittest.mock import MagicMock, create_autospec, patch

from sphinx.application import Sphinx
from sphinx.config import Config

import sphinx_autodoc_typehints as sat
from sphinx_autodoc_typehints import _inject_overload_signatures, process_docstring, process_signature
from sphinx_autodoc_typehints._resolver._util import get_obj_location
from sphinx_autodoc_typehints.patches import _OVERLOADS_CACHE


class _ClassWithPrivate:
    def __secret(self, x: int) -> str: ...


def _make_sig_app(**overrides: Any) -> Sphinx:
    defaults = {
        "typehints_fully_qualified": False,
        "simplify_optional_unions": False,
        "typehints_formatter": None,
        "typehints_use_signature": False,
        "typehints_use_signature_return": False,
        "autodoc_type_aliases": {},
    }
    defaults.update(overrides)
    config = create_autospec(Config, **defaults)  # ty: ignore[invalid-argument-type]
    return create_autospec(Sphinx, config=config)


def _make_docstring_app(**overrides: Any) -> Sphinx:
    config = MagicMock()
    config.__getitem__ = getattr
    config.autodoc_type_aliases = {}
    config.autodoc_mock_imports = []
    config.typehints_fully_qualified = False
    config.simplify_optional_unions = False
    config.typehints_formatter = None
    config.typehints_document_rtype = True
    config.typehints_document_rtype_none = True
    config.typehints_use_rtype = True
    config.typehints_defaults = None
    config.always_document_param_types = False
    config.python_display_short_literal_types = False
    for key, value in overrides.items():
        setattr(config, key, value)
    app = MagicMock(spec=Sphinx)
    app.config = config
    app.env = None
    return app


def test_process_signature_class_attr_lookup_fails() -> None:
    """Line 111: class name not found on module returns None."""

    def fake_method(self: Any, x: int) -> str: ...

    fake_method.__annotations__ = {"x": int, "return": str}
    fake_method.__qualname__ = "NonExistentClass.method"
    fake_method.__module__ = __name__

    app = _make_sig_app()
    result = process_signature(app, "method", "test.NonExistentClass.method", fake_method, MagicMock(), "", "")
    assert result is None


def test_process_signature_private_name_mangling() -> None:
    """Line 114: dunder-prefixed method names get mangled."""
    method = _ClassWithPrivate.__dict__["_ClassWithPrivate__secret"]

    app = _make_sig_app()
    result = process_signature(app, "method", "test_init._ClassWithPrivate.__secret", method, MagicMock(), "", "")
    assert result is not None
    sig_str, _ = result
    assert "self" not in sig_str


def test_process_docstring_sphinx_signature_raises_value_error() -> None:
    """Lines 157-158: sphinx_signature raising ValueError sets signature to None."""

    def func(x: int) -> str: ...

    app = _make_docstring_app(typehints_document_rtype=False)
    lines: list[str] = [":param x: the x"]
    with patch("sphinx_autodoc_typehints.sphinx_signature", side_effect=ValueError("bad")):
        process_docstring(app, "function", "func", func, None, lines)
    assert ":param x: the x" in lines


def test_process_docstring_sphinx_signature_raises_type_error() -> None:
    """Lines 157-158: sphinx_signature raising TypeError sets signature to None."""

    def func(x: int) -> str: ...

    app = _make_docstring_app(typehints_document_rtype=False)
    lines: list[str] = [":param x: the x"]
    with patch("sphinx_autodoc_typehints.sphinx_signature", side_effect=TypeError("bad")):
        process_docstring(app, "function", "func", func, None, lines)
    assert ":param x: the x" in lines


def test_inject_overload_no_qualname() -> None:
    """Line 198: obj without __qualname__ returns False."""
    obj = MagicMock(spec=[])
    obj.__module__ = "some_module"
    obj.__qualname__ = ""
    _OVERLOADS_CACHE["some_module"] = {}
    try:
        app = _make_docstring_app()
        assert _inject_overload_signatures(app, "function", "name", obj, []) is False
    finally:
        _OVERLOADS_CACHE.pop("some_module", None)


def test_inject_overload_unannotated_param_and_no_return() -> None:
    """Lines 214, 217->224: overload with unannotated param and no return annotation."""
    obj = MagicMock()
    obj.__module__ = "test_mod"
    obj.__qualname__ = "func"

    sig = inspect.Signature(
        parameters=[inspect.Parameter("x", inspect.Parameter.POSITIONAL_OR_KEYWORD)],
    )
    _OVERLOADS_CACHE["test_mod"] = {"func": [sig]}
    try:
        app = _make_docstring_app()
        lines: list[str] = []
        result = _inject_overload_signatures(app, "function", "name", obj, lines)
        assert result is True
        joined = "\n".join(lines)
        assert "**x**" in joined
        assert "\u2192" not in joined
    finally:
        _OVERLOADS_CACHE.pop("test_mod", None)


def test_inject_overload_empty_overloads_returns_false() -> None:
    """Line 233: no matching overloads returns False."""
    obj = MagicMock()
    obj.__module__ = "test_mod2"
    obj.__qualname__ = "func"

    _OVERLOADS_CACHE["test_mod2"] = {"other_func": []}
    try:
        app = _make_docstring_app()
        assert _inject_overload_signatures(app, "function", "name", obj, []) is False
    finally:
        _OVERLOADS_CACHE.pop("test_mod2", None)


def test_local_function_warning_includes_location() -> None:
    def fake_method(self: Any, x: int) -> str: ...

    fake_method.__annotations__ = {"x": int, "return": str}
    fake_method.__qualname__ = "outer.<locals>.inner"
    fake_method.__module__ = __name__

    app = _make_sig_app()
    mock_logger = MagicMock()
    with patch("sphinx_autodoc_typehints._LOGGER", mock_logger):
        process_signature(app, "method", "test.outer.<locals>.inner", fake_method, MagicMock(), "", "")
    mock_logger.warning.assert_called_once()
    kwargs = mock_logger.warning.call_args.kwargs
    assert "location" in kwargs
    assert kwargs["location"] == get_obj_location(fake_method)


def test_inject_types_no_signature() -> None:
    """Branch 261->263: signature is None skips _inject_signature."""

    def sample() -> str: ...

    app = _make_docstring_app(typehints_document_rtype=False)

    lines: list[str] = []
    sat._inject_types_to_docstring({"return": str}, None, sample, app, "function", "sample", lines)  # noqa: SLF001
    assert not any(line.startswith(":type") for line in lines)



from __future__ import annotations

from typing import TYPE_CHECKING

from viktor import AI  # module part of autodoc_mock_imports # noqa: F401

if TYPE_CHECKING:
    # Nested type guard
    from typing import Literal  # noqa: F401


class AnotherClass:
    """Another class is here"""



"""Module without deferred annotations that uses forward references."""


class Tree:
    def children(self) -> list["Tree"]:
        """
        Get child nodes.

        Returns:
            The children of this node.
        """
        return []



from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any

import attr
import attrs
import pytest
from conftest import normalize_sphinx_text

from sphinx_autodoc_typehints._resolver._attrs import backfill_attrs_annotations

if TYPE_CHECKING:
    from io import StringIO

    from sphinx.testing.util import SphinxTestApp

ATTRS_ROOT = Path(__file__).parent.parent / "roots" / "test-attrs"


@attr.s
class _ClassicAttrs:
    name = attr.ib(type=str)
    age = attr.ib(type=int)
    untyped = attr.ib()


@attr.s(auto_attribs=True)
class _AutoAttribs:
    name: str
    age: int


@attrs.define
class _ModernAttrs:
    name: str
    age: int


def test_backfill_classic_attrs() -> None:
    backfill_attrs_annotations(_ClassicAttrs)
    annotations = _ClassicAttrs.__annotations__
    assert annotations["name"] is str
    assert annotations["age"] is int
    assert "untyped" not in annotations


def test_backfill_does_not_override_existing_annotations() -> None:
    original = _AutoAttribs.__annotations__.copy()
    backfill_attrs_annotations(_AutoAttribs)
    assert _AutoAttribs.__annotations__ == original


def test_backfill_modern_attrs() -> None:
    original = _ModernAttrs.__annotations__.copy()
    backfill_attrs_annotations(_ModernAttrs)
    assert _ModernAttrs.__annotations__ == original


def test_backfill_non_attrs_class() -> None:
    class Plain:
        pass

    backfill_attrs_annotations(Plain)
    assert not hasattr(Plain, "__annotations__") or Plain.__annotations__ == {}


def test_backfill_without_attrs_installed() -> None:
    saved = sys.modules.copy()
    sys.modules["attrs"] = None  # type: ignore[assignment]
    try:

        class Dummy:
            pass

        backfill_attrs_annotations(Dummy)
    finally:
        sys.modules.update(saved)


def test_backfill_classic_attrs_creates_annotations_when_missing() -> None:
    @attr.s
    class NoAnnotations:
        x = attr.ib(type=int)

    NoAnnotations.__annotations__ = None  # type: ignore[assignment]
    backfill_attrs_annotations(NoAnnotations)
    assert NoAnnotations.__annotations__["x"] is int


@pytest.fixture
def _attrs_mod() -> Any:
    sys.path.insert(0, str(ATTRS_ROOT))
    try:
        import attrs_mod  # noqa: PLC0415  # ty: ignore[unresolved-import]

        yield attrs_mod
    finally:
        sys.path.pop(0)
        sys.modules.pop("attrs_mod", None)


@pytest.mark.sphinx("text", testroot="attrs")
def test_sphinx_build_attrs_types(app: SphinxTestApp, status: StringIO, warning: StringIO) -> None:
    template = """\
.. autoclass:: attrs_mod.ClassicAttrs
   :members:
   :undoc-members:

.. autoclass:: attrs_mod.AutoAttribs
   :members:
   :undoc-members:

.. autoclass:: attrs_mod.ModernAttrs
   :members:
   :undoc-members:
"""
    (Path(app.srcdir) / "index.rst").write_text(template)
    app.build()
    assert "build succeeded" in status.getvalue()
    result = normalize_sphinx_text((Path(app.srcdir) / "_build/text/index.txt").read_text())
    assert "str" in result
    assert "int" in result
    warn_text = warning.getvalue()
    assert "attrs_mod" not in warn_text or "forward reference" not in warn_text



"""Module demonstrating imports that are type guarded"""

from __future__ import annotations

import typing
from builtins import ValueError  # handle does not have __module__  # noqa: A004
from functools import cmp_to_key  # has __module__ but cannot get module as is builtin
from typing import TYPE_CHECKING

from demo_typing_guard_dummy import AnotherClass

if TYPE_CHECKING:
    from collections.abc import Sequence
    from decimal import Decimal

    from demo_typing_guard_dummy import Literal  # guarded by another `if TYPE_CHECKING` in demo_typing_guard_dummy


if typing.TYPE_CHECKING:
    from typing import AnyStr


if TYPE_CHECKING:  # bad import
    from functools import missing  # noqa: F401


def a[AnyStr: (bytes, str)](f: Decimal, s: AnyStr) -> Sequence[AnyStr | Decimal]:
    """
    Do.

    :param f: first
    :param s: second
    :return: result
    """
    return [f, s]


class SomeClass:
    """This class do something."""

    def create(self, item: Decimal) -> None:
        """
        Create something.

        :param item: the item in question
        """

    if TYPE_CHECKING:  # Classes doesn't have `__globals__` attribute

        def guarded(self, item: Decimal) -> None:
            """
            Guarded method.

            :param item: some item
            """


def func(_x: Literal) -> None: ...


__all__ = [
    "AnotherClass",
    "SomeClass",
    "ValueError",
    "a",
    "cmp_to_key",
]



"""Module demonstrating imports that are type guarded"""

from __future__ import annotations

import datetime

from attrs import define


@define()
class SomeClass:
    """This class does something."""

    date: datetime.date
    """Date to handle"""

    @classmethod
    def from_str(cls, input_value: str) -> SomeClass:
        """
        Initialize from string

        :param input_value: Input
        :return: result
        """
        return cls(input_value)

    @classmethod
    def from_date(cls, input_value: datetime.date) -> SomeClass:
        """
        Initialize from date

        :param input_value: Input
        :return: result
        """
        return cls(input_value)

    @classmethod
    def from_time(cls, input_value: datetime.time) -> SomeClass:
        """
        Initialize from time

        :param input_value: Input
        :return: result
        """
        return cls(input_value)

    def calculate_thing(self, number: float) -> datetime.timedelta:  # noqa: PLR6301
        """
        Calculate a thing

        :param number: Input
        :return: result
        """
        return datetime.timedelta(number)


__all__ = ["SomeClass"]



from __future__ import annotations

from dataclasses import dataclass
from typing import NamedTuple


def undocumented_function(x: int) -> str:
    """Hi"""

    return str(x)


def undocumented_function_with_defaults(x: int, y: str = "hello") -> str:
    """Hi"""

    return str(x) + y


class MyNamedTuple(NamedTuple):
    """A named tuple."""

    x: int
    y: str = "hello"


@dataclass
class DataClass:
    """Class docstring."""

    x: int



from __future__ import annotations

import sys
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from io import StringIO

    from sphinx.testing.util import SphinxTestApp


class MyClass:
    """A class with a property."""

    @property
    def n_unique_nonzero(self) -> int:
        """Number of unique nonzero elements."""
        return 0

    def regular_method(self, x: int) -> int:  # noqa: PLR6301
        """Do something.

        Args:
            x: a number
        """
        return x


@pytest.mark.sphinx("text", testroot="integration")
def test_method_lookup_does_not_crash(
    app: SphinxTestApp,
    status: StringIO,
    warning: StringIO,  # noqa: ARG001
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    (Path(app.srcdir) / "index.rst").write_text(
        dedent("""\
        Test
        ====

        .. autoclass:: mod.MyClass
           :members:
    """)
    )
    monkeypatch.setitem(sys.modules, "mod", sys.modules[__name__])
    app.build()
    assert "build succeeded" in status.getvalue()



from __future__ import annotations

from typing import Annotated

from typing_extensions import Doc


def greet(
    name: Annotated[str, Doc("The person's name")], greeting: Annotated[str, Doc("The greeting phrase")]
) -> Annotated[str, Doc("The full greeting message")]:
    """Say hello."""
    return f"{greeting}, {name}!"


def partial_doc(x: Annotated[int, Doc("The x value")], y: int) -> int:
    """Compute sum.

    :param y: The y value
    """
    return x + y


def no_doc(x: Annotated[int, 42]) -> int:
    """Identity."""
    return x


def docstring_wins(x: Annotated[int, Doc("Doc description")]) -> int:
    """Override.

    :param x: Docstring description
    """
    return x



from __future__ import annotations

import sys
import types
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from io import StringIO

    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx("text", testroot="integration")
def test_guarded_import_missing_name_no_warning(
    app: SphinxTestApp,
    status: StringIO,
    warning: StringIO,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    target_mod = types.ModuleType("target_mod")
    target_mod.__file__ = "/fake/target_mod.py"

    source = dedent("""\
    from __future__ import annotations
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from target_mod import nonexistent_name

    def func(x: int) -> int:
        '''Do something.

        Args:
            x: a number
        '''
        return x
    """)
    user_mod = types.ModuleType("user_mod")
    user_mod.__file__ = "/fake/user_mod.py"
    exec(compile(source, "/fake/user_mod.py", "exec"), user_mod.__dict__)  # noqa: S102

    monkeypatch.setitem(sys.modules, "target_mod", target_mod)
    monkeypatch.setitem(sys.modules, "user_mod", user_mod)

    (Path(app.srcdir) / "index.rst").write_text(
        dedent("""\
        Test
        ====

        .. autofunction:: user_mod.func
    """)
    )
    app.build()
    assert "build succeeded" in status.getvalue()
    assert "Failed guarded type import" not in warning.getvalue()



"""Module with no type annotations, simulating a C extension."""


def greet(name, greeting):
    return f"{greeting}, {name}!"


class Calculator:
    value = 0

    def add(self, x):
        self.value += x
        return self

    class Inner:
        def process(self, data):
            return data


async def fetch(url):
    return url



from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from conftest import normalize_sphinx_text

numpydoc = pytest.importorskip("numpydoc")

if TYPE_CHECKING:
    from io import StringIO

    from sphinx.testing.util import SphinxTestApp


def simple_func(x: int, y: str) -> str:
    """
    Do something simple.

    Parameters
    ----------
    x : int
        The x value.
    y : str
        The y value.

    Returns
    -------
    result : str
        The combined result.
    """
    return f"{x}{y}"  # pragma: no cover


def raises_func(x: int) -> None:
    """
    Raise on bad input.

    Parameters
    ----------
    x : int
        The input value.

    Raises
    ------
    ValueError
        If x is negative.
    TypeError
        If x is not an integer.
    """


def multi_return(x: int) -> tuple[str, int]:
    """
    Return multiple values.

    Parameters
    ----------
    x : int
        The input.

    Returns
    -------
    name : str
        The name.
    value : int
        The value.
    """
    return str(x), x  # pragma: no cover


def no_params() -> str:
    """
    Function with no parameters.

    Returns
    -------
    str
        A greeting.
    """
    return "hello"  # pragma: no cover


def _build_and_get_output(
    app: SphinxTestApp,
    status: StringIO,
    monkeypatch: pytest.MonkeyPatch,
    func_name: str,
) -> str:
    template = f"""\
.. autofunction:: mod_numpy.{func_name}
"""
    (Path(app.srcdir) / "index.rst").write_text(template)
    monkeypatch.setitem(sys.modules, "mod_numpy", sys.modules[__name__])
    app.build()
    assert "build succeeded" in status.getvalue()
    return normalize_sphinx_text((Path(app.srcdir) / "_build/text/index.txt").read_text())


@pytest.mark.sphinx("text", testroot="issue_347")
def test_simple_params_and_return(
    app: SphinxTestApp,
    status: StringIO,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    result = _build_and_get_output(app, status, monkeypatch, "simple_func")
    assert "Parameters:" in result
    assert "**x** (" in result
    assert "**y** (" in result
    assert "Return type:" in result


@pytest.mark.sphinx("text", testroot="issue_347")
def test_raises_section(
    app: SphinxTestApp,
    status: StringIO,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    result = _build_and_get_output(app, status, monkeypatch, "raises_func")
    assert "Parameters:" in result
    assert "Raises:" in result
    assert "ValueError" in result
    assert "TypeError" in result


@pytest.mark.sphinx("text", testroot="issue_347")
def test_multi_return(
    app: SphinxTestApp,
    status: StringIO,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    result = _build_and_get_output(app, status, monkeypatch, "multi_return")
    assert "Return type:" in result


@pytest.mark.sphinx("text", testroot="issue_347")
def test_no_params(
    app: SphinxTestApp,
    status: StringIO,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    result = _build_and_get_output(app, status, monkeypatch, "no_params")
    assert "Return type:" in result



from __future__ import annotations

import sys
import types
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from io import StringIO

    from sphinx.testing.util import SphinxTestApp

_mod_pep695 = types.ModuleType("_mod_pep695")
_mod_pep695.__file__ = __file__
exec(  # noqa: S102
    dedent("""\
    from __future__ import annotations

    class Foo[T]:
        \"\"\"A generic class.\"\"\"

        def __init__(self, thing: T) -> None:
            \"\"\"Init.

            :param thing: the thing
            \"\"\"

        def get(self) -> T:
            \"\"\"Get the thing.

            :return: the thing
            \"\"\"
            ...

    class Multi[K, V]:
        \"\"\"A class with multiple type params.\"\"\"

        def lookup(self, key: K) -> V:
            \"\"\"Look up.

            :param key: the key
            \"\"\"
            ...

    def identity[U](x: U) -> U:
        \"\"\"Identity function.

        :param x: input
        \"\"\"
        return x
    """),
    _mod_pep695.__dict__,
)


@pytest.mark.sphinx("text", testroot="integration")
def test_pep695_class_type_params(
    app: SphinxTestApp, status: StringIO, warning: StringIO, monkeypatch: pytest.MonkeyPatch
) -> None:
    (Path(app.srcdir) / "index.rst").write_text(
        dedent("""\
        Test
        ====

        .. autoclass:: mod.Foo
           :members:
    """)
    )
    monkeypatch.setitem(sys.modules, "mod", _mod_pep695)
    app.build()
    assert "build succeeded" in status.getvalue()
    assert "Cannot resolve forward reference" not in warning.getvalue()


@pytest.mark.sphinx("text", testroot="integration")
def test_pep695_class_multiple_type_params(
    app: SphinxTestApp, status: StringIO, warning: StringIO, monkeypatch: pytest.MonkeyPatch
) -> None:
    (Path(app.srcdir) / "index.rst").write_text(
        dedent("""\
        Test
        ====

        .. autoclass:: mod.Multi
           :members:
    """)
    )
    monkeypatch.setitem(sys.modules, "mod", _mod_pep695)
    app.build()
    assert "build succeeded" in status.getvalue()
    assert "Cannot resolve forward reference" not in warning.getvalue()


@pytest.mark.sphinx("text", testroot="integration")
def test_pep695_function_type_params(
    app: SphinxTestApp, status: StringIO, warning: StringIO, monkeypatch: pytest.MonkeyPatch
) -> None:
    (Path(app.srcdir) / "index.rst").write_text(
        dedent("""\
        Test
        ====

        .. autofunction:: mod.identity
    """)
    )
    monkeypatch.setitem(sys.modules, "mod", _mod_pep695)
    app.build()
    assert "build succeeded" in status.getvalue()
    assert "Cannot resolve forward reference" not in warning.getvalue()



def greet(name: str, greeting: str) -> str: ...

class Calculator:
    value: int
    def add(self, x: int) -> Calculator: ...
    class Inner:
        def process(self, data: bytes) -> bytes: ...

async def fetch(url: str) -> str: ...



from __future__ import annotations

from typing import Annotated

from typing_extensions import Doc


def compute(
    x: Annotated[int, Doc("The x input")], y: Annotated[int, Doc("The y input")]
) -> Annotated[int, Doc("The sum")]:
    """
    Compute the sum.

    Parameters
    ----------
    x
        Placeholder.
    y
        Placeholder.
    """
    return x + y


def transform(data: Annotated[str, Doc("The input data")]) -> Annotated[str, Doc("The transformed result")]:
    """Transform data."""
    return data.upper()



from __future__ import annotations

import re
import sys
from dataclasses import dataclass
from inspect import isclass
from pathlib import Path
from textwrap import dedent, indent
from typing import (  # no type comments
    TYPE_CHECKING,
    Any,
    Literal,
    NewType,
    Optional,
    TypeVar,
    Union,
    overload,
)

import pytest
from conftest import normalize_sphinx_text

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Callable
    from io import StringIO
    from mailbox import Mailbox
    from types import CodeType, ModuleType

    from sphinx.testing.util import SphinxTestApp

T = TypeVar("T")
W = NewType("W", str)


@dataclass
class WarningInfo:
    """Properties and assertion methods for warnings."""

    regexp: str
    type: str

    def assert_regexp(self, message: str) -> None:
        regexp = self.regexp
        msg = f"Regex pattern did not match.\n Regex: {regexp!r}\n Input: {message!r}"
        assert re.search(regexp, message), msg

    def assert_type(self, message: str) -> None:
        expected = f"[{self.type}]"
        msg = f"Warning did not contain type and subtype.\n Expected: {expected}\n Input: {message}"
        assert expected in message, msg

    def assert_warning(self, message: str) -> None:
        self.assert_regexp(message)
        self.assert_type(message)


def expected(expected: str, **options: Any) -> Callable[[T], T]:
    def dec(val: T) -> T:
        val.EXPECTED = expected  # ty: ignore[unresolved-attribute]
        val.OPTIONS = options  # ty: ignore[unresolved-attribute]
        return val

    return dec


def warns(info: WarningInfo) -> Callable[[T], T]:
    def dec(val: T) -> T:
        val.WARNING = info  # ty: ignore[unresolved-attribute]
        return val

    return dec


@expected("mod.get_local_function()")
def get_local_function():  # noqa: ANN201
    def wrapper(self) -> str:  # noqa: ANN001
        """
        Wrapper
        """

    return wrapper


@warns(WarningInfo(regexp="Cannot handle as a local function", type="sphinx_autodoc_typehints.local_function"))
@expected(
    """\
class mod.Class(x, y, z=None)

   Initializer docstring.

   Parameters:
      * **x** ("bool") -- foo

      * **y** ("int") -- bar

      * **z** ("str" | "None") -- baz

   class InnerClass

      Inner class.

      inner_method(x)

         Inner method.

         Parameters:
            **x** ("bool") -- foo

         Return type:
            "str"

   classmethod a_classmethod(x, y, z=None)

      Classmethod docstring.

      Parameters:
         * **x** ("bool") -- foo

         * **y** ("int") -- bar

         * **z** ("str" | "None") -- baz

      Return type:
         "str"

   a_method(x, y, z=None)

      Method docstring.

      Parameters:
         * **x** ("bool") -- foo

         * **y** ("int") -- bar

         * **z** ("str" | "None") -- baz

      Return type:
         "str"

   property a_property: str

      Property docstring

   static a_staticmethod(x, y, z=None)

      Staticmethod docstring.

      Parameters:
         * **x** ("bool") -- foo

         * **y** ("int") -- bar

         * **z** ("str" | "None") -- baz

      Return type:
         "str"

   locally_defined_callable_field() -> str

      Wrapper

      Return type:
         "str"
    """,
)
class Class:
    """
    Initializer docstring.

    :param x: foo
    :param y: bar
    :param z: baz
    """

    def __init__(self, x: bool, y: int, z: Optional[str] = None) -> None:
        pass

    def a_method(self, x: bool, y: int, z: Optional[str] = None) -> str:
        """
        Method docstring.

        :param x: foo
        :param y: bar
        :param z: baz
        """

    def _private_method(self, x: str) -> str:
        """
        Private method docstring.

        :param x: foo
        """

    def __dunder_method(self, x: str) -> str:
        """
        Dunder method docstring.

        :param x: foo
        """

    def __magic_custom_method__(self, x: str) -> str:  # noqa: PLW3201
        """
        Magic dunder method docstring.

        :param x: foo
        """

    @classmethod
    def a_classmethod(cls, x: bool, y: int, z: Optional[str] = None) -> str:
        """
        Classmethod docstring.

        :param x: foo
        :param y: bar
        :param z: baz
        """

    @staticmethod
    def a_staticmethod(x: bool, y: int, z: Optional[str] = None) -> str:
        """
        Staticmethod docstring.

        :param x: foo
        :param y: bar
        :param z: baz
        """

    @property
    def a_property(self) -> str:
        """
        Property docstring
        """

    class InnerClass:
        """
        Inner class.
        """

        def inner_method(self, x: bool) -> str:
            """
            Inner method.

            :param x: foo
            """

        def __dunder_inner_method(self, x: bool) -> str:
            """
            Dunder inner method.

            :param x: foo
            """

    locally_defined_callable_field = get_local_function()


@expected(
    """\
exception mod.DummyException(message)

   Exception docstring

   Parameters:
      **message** ("str") -- blah
""",
)
class DummyException(Exception):  # noqa: N818
    """
    Exception docstring

    :param message: blah
    """

    def __init__(self, message: str) -> None:
        super().__init__(message)


@expected(
    """\
mod.function(x, y, z_=None)

   Function docstring.

   Parameters:
      * **x** ("bool") -- foo

      * **y** ("int") -- bar

      * **z_** ("str" | "None") -- baz

   Returns:
      something

   Return type:
      bytes
""",
)
def function(x: bool, y: int, z_: Optional[str] = None) -> str:
    """
    Function docstring.

    :param x: foo
    :param y: bar
    :param z\\_: baz
    :return: something
    :rtype: bytes
    """


@expected(
    """\
mod.function_with_starred_documentation_param_names(*args, **kwargs)

   Function docstring.

   Usage:

      print(1)

   Parameters:
      * ***args** ("int") -- foo

      * ****kwargs** ("str") -- bar
""",
)
def function_with_starred_documentation_param_names(*args: int, **kwargs: str):  # noqa: ANN201
    r"""
    Function docstring.

    Usage::

        print(1)

    :param \*args: foo
    :param \**kwargs: bar
    """


@expected(
    """\
mod.function_with_escaped_default(x='\\\\x08')

   Function docstring.

   Parameters:
      **x** ("str") -- foo
""",
)
def function_with_escaped_default(x: str = "\b"):  # noqa: ANN201
    """
    Function docstring.

    :param x: foo
    """


@warns(
    WarningInfo(
        regexp="Cannot resolve forward reference in type annotations", type="sphinx_autodoc_typehints.forward_reference"
    )
)
@expected(
    """\
mod.function_with_unresolvable_annotation(x)

   Function docstring.

   Parameters:
      **x** (a.b.c) -- foo
""",
)
def function_with_unresolvable_annotation(x: a.b.c):  # noqa: ANN201, F821  # ty: ignore[unresolved-reference]
    """
    Function docstring.

    :arg x: foo
    """


@expected(
    """\
mod.function_with_typehint_comment(x, y)

   Function docstring.

   Parameters:
      * **x** ("int") -- foo

      * **y** ("str") -- bar

   Return type:
      "None"
""",
)
def function_with_typehint_comment(  # noqa: ANN201
    x,  # type: int  # noqa: ANN001
    y,  # type: str  # noqa: ANN001
):
    # type: (...) -> None
    """
    Function docstring.

    :parameter x: foo
    :parameter y: bar
    """


@expected(
    """\
class mod.ClassWithTypehints(x)

   Class docstring.

   Parameters:
      **x** ("int") -- foo

   foo(x)

      Method docstring.

      Parameters:
         **x** ("str") -- foo

      Return type:
         "int"

   method_without_typehint(x)

      Method docstring.
""",
)
class ClassWithTypehints:
    """
    Class docstring.

    :param x: foo
    """

    def __init__(
        self,
        x,  # type: int  # noqa: ANN001
    ) -> None:
        # type: (...) -> None
        pass

    def foo(  # noqa: ANN201, PLR6301
        self,
        x,  # type: str  # noqa: ANN001, ARG002
    ):
        # type: (...) -> int
        """
        Method docstring.

        :arg x: foo
        """
        return 42

    def method_without_typehint(self, x):  # noqa: ANN001, ANN201, ARG002, PLR6301
        """
        Method docstring.
        """
        # test that multiline str can be correctly indented
        multiline_str = """
test
"""
        return multiline_str  # noqa: RET504


@expected(
    """\
mod.function_with_typehint_comment_not_inline(x=None, *y, z, **kwargs)

   Function docstring.

   Parameters:
      * **x** ("str" | "bytes" | "None") -- foo

      * **y** ("str") -- bar

      * **z** ("bytes") -- baz

      * **kwargs** ("int") -- some kwargs

   Return type:
      "None"
""",
)
def function_with_typehint_comment_not_inline(x=None, *y, z, **kwargs):  # noqa: ANN001, ANN002, ANN003, ANN201
    # type: (Union[str, bytes, None], *str, bytes, **int) -> None
    """
    Function docstring.

    :arg x: foo
    :argument y: bar
    :parameter z: baz
    :parameter kwargs: some kwargs
    """


@expected(
    """\
class mod.ClassWithTypehintsNotInline(x=None)

   Class docstring.

   Parameters:
      **x** ("Callable"[["int", "bytes"], "int"] | "None") -- foo

   foo(x=1)

      Method docstring.

      Parameters:
         **x** ("Callable"[["int", "bytes"], "int"]) -- foo

      Return type:
         "int"

   classmethod mk(x=None)

      Method docstring.

      Parameters:
         **x** ("Callable"[["int", "bytes"], "int"] | "None") -- foo

      Return type:
         "ClassWithTypehintsNotInline"
""",
)
class ClassWithTypehintsNotInline:
    """
    Class docstring.

    :param x: foo
    """

    def __init__(self, x=None) -> None:  # type: (Optional[Callable[[int, bytes], int]]) -> None  # noqa: ANN001
        pass

    def foo(self, x=1):  # type: (Callable[[int, bytes], int]) -> int  # noqa: ANN001, ANN201, PLR6301
        """
        Method docstring.

        :param x: foo
        """
        return x(1, b"")  # ty: ignore[call-non-callable]

    @classmethod
    def mk(  # noqa: ANN206
        cls,
        x=None,  # noqa: ANN001
    ):  # type: (Optional[Callable[[int, bytes], int]]) -> ClassWithTypehintsNotInline
        """
        Method docstring.

        :param x: foo
        """
        return cls(x)


@expected(
    """\
mod.undocumented_function(x)

   Hi

   Return type:
      "str"
""",
)
def undocumented_function(x: int) -> str:
    """Hi"""

    return str(x)


@expected(
    """\
class mod.DataClass(x)

   Class docstring.
""",
)
@dataclass
class DataClass:
    """Class docstring."""

    x: int


@expected(
    """\
class mod.Decorator(func)

   Initializer docstring.

   Parameters:
      **func** ("Callable"[["int", "str"], "str"]) -- function
""",
)
class Decorator:
    """
    Initializer docstring.

    :param func: function
    """

    def __init__(self, func: Callable[[int, str], str]) -> None:
        pass


@expected(
    """\
mod.mocked_import(x)

   A docstring.

   Parameters:
      **x** ("Mailbox") -- function
""",
)
def mocked_import(x: Mailbox):  # noqa: ANN201
    """
    A docstring.

    :param x: function
    """


@expected(
    """\
mod.func_with_examples()

   A docstring.

   Return type:
      "int"

   -[ Examples ]-

   Here are a couple of examples of how to use this function.
""",
)
def func_with_examples() -> int:
    """
    A docstring.

    .. rubric:: Examples

    Here are a couple of examples of how to use this function.
    """


@overload
def func_with_overload(a: int, b: int) -> None: ...


@overload
def func_with_overload(a: str, b: str) -> None: ...


@expected(
    """\
mod.func_with_overload(a, b)

   Overloads:
      * **a** (int), **b** (int) → None

      * **a** (str), **b** (str) → None

   f does the thing. The arguments can either be ints or strings but
   they must both have the same type.

   Parameters:
      * **a** ("int" | "str") -- The first thing

      * **b** ("int" | "str") -- The second thing
""",
)
def func_with_overload(a: Union[int, str], b: Union[int, str]) -> None:
    """
    f does the thing. The arguments can either be ints or strings but they must
    both have the same type.

    Parameters
    ----------
    a:
        The first thing
    b:
        The second thing
    """


@overload
def overload_with_complex_types(x: list[int]) -> dict[str, int]: ...


@overload
def overload_with_complex_types(x: dict[str, int]) -> list[int]: ...


@expected(
    """\
mod.overload_with_complex_types(x)

   Overloads:
      * **x** (list[int]) → dict[str, int]

      * **x** (dict[str, int]) → list[int]

   Function with overloaded complex types.

   Parameters:
      **x** ("list"["int"] | "dict"["str", "int"]) -- Input value
""",
)
def overload_with_complex_types(x: list[int] | dict[str, int]) -> dict[str, int] | list[int]:
    """Function with overloaded complex types.

    Parameters:
        x: Input value
    """
    if isinstance(x, list):
        return {str(i): i for i in x}
    return list(x.values())


@expected(
    """\
mod.func_literals_long_format(a, b)

   A docstring.

   Parameters:
      * **a** ("Literal"["'arg1'", "'arg2'"]) -- Argument that can
        take either of two literal values.

      * **b** ("Literal"["'arg1'", "'arg2'"]) -- Argument that can
        take either of two literal values.

   Return type:
      "None"
""",
)
def func_literals_long_format(a: Literal["arg1", "arg2"], b: Literal["arg1", "arg2"]) -> None:
    """
    A docstring.

    :param a: Argument that can take either of two literal values.
    :param b: Argument that can take either of two literal values.
    """


@expected(
    """\
mod.func_literals_short_format(a, b)

   A docstring.

   Parameters:
      * **a** ("'arg1'" | "'arg2'") -- Argument that can take either
        of two literal values.

      * **b** ("'arg1'" | "'arg2'") -- Argument that can take either
        of two literal values.

   Return type:
      "None"
""",
    python_display_short_literal_types=True,
)
def func_literals_short_format(a: Literal["arg1", "arg2"], b: Literal["arg1", "arg2"]) -> None:
    """
    A docstring.

    :param a: Argument that can take either of two literal values.
    :param b: Argument that can take either of two literal values.
    """


@expected(
    """\
class mod.TestClassAttributeDocs

   A class

   code: CodeType | None

      An attribute
""",
)
class TestClassAttributeDocs:
    """A class"""

    code: Optional[CodeType]
    """An attribute"""


@expected(
    """\
mod.func_with_examples_and_returns_after()

   f does the thing.

   -[ Examples ]-

   Here is an example

   Return type:
      "int"

   Returns:
      The index of the widget
""",
)
def func_with_examples_and_returns_after() -> int:
    """
    f does the thing.

    Examples
    --------

    Here is an example

    :returns: The index of the widget
    """


@expected(
    """\
mod.func_with_parameters_and_stuff_after(a, b)

   A func

   Parameters:
      * **a** ("int") -- a tells us something

      * **b** ("int") -- b tells us something

   Return type:
      "int"

   More info about the function here.
""",
)
def func_with_parameters_and_stuff_after(a: int, b: int) -> int:
    """A func

    :param a: a tells us something
    :param b: b tells us something

    More info about the function here.
    """


@expected(
    """\
mod.func_with_rtype_in_weird_spot(a, b)

   A func

   Parameters:
      * **a** ("int") -- a tells us something

      * **b** ("int") -- b tells us something

   -[ Examples ]-

   Here is an example

   Returns:
      The index of the widget

   More info about the function here.

   Return type:
      int
""",
)
def func_with_rtype_in_weird_spot(a: int, b: int) -> int:
    """A func

    :param a: a tells us something
    :param b: b tells us something

    Examples
    --------

    Here is an example

    :returns: The index of the widget

    More info about the function here.

    :rtype: int
    """


@expected(
    """\
mod.empty_line_between_parameters(a, b)

   A func

   Parameters:
      * **a** ("int") --

        One of the following possibilities:

        * a

        * b

        * c

      * **b** ("int") --

        Whatever else we have to say.

        There is more of it And here too

   Return type:
      "int"

   More stuff here.
""",
)
def empty_line_between_parameters(a: int, b: int) -> int:
    """A func

    :param a: One of the following possibilities:

             - a

             - b

             - c

    :param b: Whatever else we have to say.

              There is more of it And here too

    More stuff here.
    """


@expected(
    """\
mod.func_with_code_block()

   A docstring.

   You would say:

      print("some python code here")

   Return type:
      "int"

   -[ Examples ]-

   Here are a couple of examples of how to use this function.
""",
)
def func_with_code_block() -> int:
    """
    A docstring.

    You would say:

    .. code-block::

        print("some python code here")


    .. rubric:: Examples

    Here are a couple of examples of how to use this function.
    """


@expected(
    """
    mod.func_with_definition_list()

       Some text and then a definition list.

       Return type:
          "int"

       abc
          x

       xyz
          something
    """,
)
def func_with_definition_list() -> int:
    """Some text and then a definition list.

    abc
        x

    xyz
        something
    """
    # See https://github.com/tox-dev/sphinx-autodoc-typehints/issues/302


@expected(
    """\
mod.decorator_2(f)

   Run the decorated function with *asyncio.run*.

   Parameters:
      **f** ("Any") -- The function to wrap.

   Return type:
      "Any"

   -[ Examples ]-

      A
""",
)
def decorator_2(f: Any) -> Any:
    """Run the decorated function with `asyncio.run`.

    Parameters
    ----------
    f
        The function to wrap.

    Examples
    --------

    .. code-block:: python

        A
    """
    assert f is not None


@expected(
    """
    class mod.ParamAndAttributeHaveSameName(blah)

       A Class

       Parameters:
          **blah** ("CodeType") -- Description of parameter blah

       blah: ModuleType

          Description of attribute blah

    """,
)
class ParamAndAttributeHaveSameName:
    """
    A Class

    Parameters
    ----------
    blah:
        Description of parameter blah
    """

    def __init__(self, blah: CodeType) -> None:
        pass

    blah: ModuleType
    """Description of attribute blah"""


@expected(
    """
    mod.napoleon_returns()

       A function.

       Return type:
          "CodeType"

       Returns:
          The info about the whatever.
    """,
)
def napoleon_returns() -> CodeType:
    """
    A function.

    Returns
    -------
        The info about the whatever.
    """


@expected(
    """
    mod.google_docstrings(arg1, arg2)

       Summary line.

       Extended description of function.

       Parameters:
          * **arg1** ("CodeType") -- Description of arg1

          * **arg2** ("ModuleType") -- Description of arg2

       Return type:
          "CodeType"

       Returns:
          Description of return value

    """,
)
def google_docstrings(arg1: CodeType, arg2: ModuleType) -> CodeType:
    """Summary line.

    Extended description of function.

    Args:
        arg1: Description of arg1
        arg2: Description of arg2

    Returns:
        Description of return value
    """


@expected(
    """
    mod.docstring_with_multiline_note_after_params(param)

       Do something.

       Parameters:
          **param** ("int") -- A parameter.

       Return type:
          "None"

       Note:

         Some notes. More notes

    """,
)
def docstring_with_multiline_note_after_params(param: int) -> None:
    """Do something.

    Args:
        param: A parameter.

    Note:

        Some notes.
        More notes
    """


@expected(
    """
    mod.docstring_with_bullet_list_after_params(param)

       Do something.

       Parameters:
          **param** ("int") -- A parameter.

       Return type:
          "None"

       * A: B

       * C: D

    """,
)
def docstring_with_bullet_list_after_params(param: int) -> None:
    """Do something.

    Args:
        param: A parameter.

    * A: B
    * C: D
    """


@expected(
    """
    mod.docstring_with_definition_list_after_params(param)

       Do something.

       Parameters:
          **param** ("int") -- A parameter.

       Return type:
          "None"

       Term
          A description

          maybe multiple lines

       Next Term
          Something about it

    """,
)
def docstring_with_definition_list_after_params(param: int) -> None:
    """Do something.

    Args:
        param: A parameter.

    Term
        A description

        maybe multiple lines

    Next Term
        Something about it
    """


@expected(
    """
    mod.docstring_with_enum_list_after_params(param)

       Do something.

       Parameters:
          **param** ("int") -- A parameter.

       Return type:
          "None"

       1. A: B

       2. C: D

    """,
)
def docstring_with_enum_list_after_params(param: int) -> None:
    """Do something.

    Args:
        param: A parameter.

    1. A: B
    2. C: D
    """


@warns(WarningInfo(regexp="Definition list ends without a blank line", type="docutils"))
@expected(
    """
    mod.docstring_with_definition_list_after_params_no_blank_line(param)

       Do something.

       Parameters:
          **param** ("int") -- A parameter.

       Return type:
          "None"

       Term
          A description

          maybe multiple lines

       Next Term
          Something about it

       -[ Example ]-
    """,
)
def docstring_with_definition_list_after_params_no_blank_line(param: int) -> None:
    """Do something.

    Args:
        param: A parameter.

    Term
        A description

        maybe multiple lines

    Next Term
        Something about it
    .. rubric:: Example
    """


@expected(
    """
    mod.has_typevar(param)

       Do something.

       Parameters:
          **param** ("TypeVar"("T")) -- A parameter.

       Return type:
          "TypeVar"("T")

    """,
)
def has_typevar[T](param: T) -> T:
    """Do something.

    Args:
        param: A parameter.
    """
    return param


@expected(
    """
    mod.has_newtype(param)

       Do something.

       Parameters:
          **param** ("W" ("str")) -- A parameter.

       Return type:
          "W" ("str")

    """,
)
def has_newtype(param: W) -> W:
    """Do something.

    Args:
        param: A parameter.
    """
    return param


AUTO_FUNCTION = ".. autofunction:: mod.{}"
AUTO_CLASS = """\
.. autoclass:: mod.{}
    :members:
"""
AUTO_EXCEPTION = """\
.. autoexception:: mod.{}
   :members:
"""


@expected(
    """
    mod.typehints_use_signature(a: AsyncGenerator) -> AsyncGenerator

       Do something.

       Parameters:
          **a** ("AsyncGenerator") -- blah

       Return type:
          "AsyncGenerator"

    """,
    typehints_use_signature=True,
    typehints_use_signature_return=True,
)
def typehints_use_signature(a: AsyncGenerator) -> AsyncGenerator:
    """Do something.

    Args:
        a: blah
    """
    return a


@expected(
    """
    mod.typehints_no_rtype_none()

       Do something.

    """,
    typehints_document_rtype_none=False,
)
def typehints_no_rtype_none() -> None:
    """Do something."""


prolog = """
.. |test_node_start| replace:: {test_node_start}
""".format(test_node_start="test_start")


@expected(
    """
    mod.docstring_with_multiline_note_after_params_prolog_replace(param)

       Do something.

       Parameters:
          **param** ("int") -- A parameter.

       Return type:
          "None"

       Note:

         Some notes. test_start More notes

    """,
    rst_prolog=prolog,
)
def docstring_with_multiline_note_after_params_prolog_replace(param: int) -> None:
    """Do something.

    Args:
        param: A parameter.

    Note:

        Some notes. |test_node_start|
        More notes
    """


epilog = """
.. |test_node_end| replace:: {test_node_end}
""".format(test_node_end="test_end")


@expected(
    """
    mod.docstring_with_multiline_note_after_params_epilog_replace(param)

       Do something.

       Parameters:
          **param** ("int") -- A parameter.

       Return type:
          "None"

       Note:

         Some notes. test_end More notes

    """,
    rst_epilog=epilog,
)
def docstring_with_multiline_note_after_params_epilog_replace(param: int) -> None:
    """Do something.

    Args:
        param: A parameter.

    Note:

        Some notes. |test_node_end|
        More notes
    """


@expected(
    """
    mod.docstring_with_see_also()

       Test

       See also: more info at <https://example.com>`_.

       Return type:
          "str"

    """
)
def docstring_with_see_also() -> str:
    """
    Test

    .. seealso:: more info at <https://example.com>`_.
    """
    return ""


@expected(
    """
    mod.has_doctest1()

       Test that we place the return type correctly when the function has
       a doctest.

       Return type:
          "None"

       >>> this is a fake doctest
       a
       >>> more doctest
       b
    """
)
def has_doctest1() -> None:
    r"""Test that we place the return type correctly when the function has a doctest.

    >>> this is a fake doctest
    a
    >>> more doctest
    b
    """


Unformatted = TypeVar("Unformatted")


@warns(WarningInfo(regexp="cannot cache unpickleable configuration value: 'typehints_formatter'", type="config.cache"))
@expected(
    """
    mod.typehints_formatter_applied_to_signature(param: Formatted) -> Formatted

       Do nothing

       Parameters:
          **param** (Formatted) -- A parameter

       Return type:
          Formatted

       Returns:
          The return value
    """,
    typehints_use_signature=True,
    typehints_use_signature_return=True,
    typehints_formatter=lambda _, __=None: "Formatted",
)
def typehints_formatter_applied_to_signature[Unformatted](param: Unformatted) -> Unformatted:
    """
    Do nothing

    Args:
        param: A parameter

    Returns:
        The return value
    """
    return param


# Config settings for each test run.
# Config Name: Sphinx Options as Dict.
configs = {
    "default_conf": {},
    "prolog_conf": {"rst_prolog": prolog},
    "epilog_conf": {
        "rst_epilog": epilog,
    },
    "bothlog_conf": {
        "rst_prolog": prolog,
        "rst_epilog": epilog,
    },
}


@pytest.mark.parametrize("val", [x for x in globals().values() if hasattr(x, "EXPECTED")])
@pytest.mark.parametrize("conf_run", ["default_conf", "prolog_conf", "epilog_conf", "bothlog_conf"])
@pytest.mark.sphinx("text", testroot="integration")
def test_integration(
    app: SphinxTestApp, status: StringIO, warning: StringIO, monkeypatch: pytest.MonkeyPatch, val: Any, conf_run: str
) -> None:
    if isclass(val) and issubclass(val, BaseException):
        template = AUTO_EXCEPTION
    elif isclass(val):
        template = AUTO_CLASS
    else:
        template = AUTO_FUNCTION

    (Path(app.srcdir) / "index.rst").write_text(template.format(val.__name__))
    app.config.__dict__.update(configs[conf_run])
    app.config.__dict__.update(val.OPTIONS)
    app.config.always_use_bars_union = True
    monkeypatch.setitem(sys.modules, "mod", sys.modules[__name__])
    app.build()
    assert "build succeeded" in status.getvalue()  # Build succeeded

    warning_info: Union[WarningInfo, None] = getattr(val, "WARNING", None)
    value = warning.getvalue().strip()
    if warning_info:
        warning_info.assert_warning(value)
    else:
        assert not value

    result = normalize_sphinx_text((Path(app.srcdir) / "_build/text/index.txt").read_text())

    expected = normalize_sphinx_text(val.EXPECTED)
    try:
        assert result.strip() == dedent(expected).strip()
    except Exception:
        indented = indent(f'"""\n{result}\n"""', " " * 4)
        print(f"@expected(\n{indented}\n)\n")  # noqa: T201
        raise



from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest
from conftest import normalize_sphinx_text

from sphinx_autodoc_typehints._resolver._type_hints import _get_type_hint

if sys.version_info >= (3, 14):
    import annotationlib

if TYPE_CHECKING:
    from io import StringIO

    from sphinx.testing.util import SphinxTestApp

SKIP_REASON = "annotationlib requires Python 3.14+"
needs_314 = pytest.mark.skipif(sys.version_info < (3, 14), reason=SKIP_REASON)


@needs_314
@pytest.mark.sphinx("text", testroot="issue_572")
def test_forward_ref_builds_without_errors(
    app: SphinxTestApp,
    status: StringIO,
    warning: StringIO,  # noqa: ARG001
) -> None:
    """Forward-referencing module builds cleanly on 3.14+ using annotationlib."""
    template = """\
.. autoclass:: mod_forward_ref.Tree
   :members:
"""
    (Path(app.srcdir) / "index.rst").write_text(template)
    app.build()
    assert "build succeeded" in status.getvalue()
    result = normalize_sphinx_text((Path(app.srcdir) / "_build/text/index.txt").read_text())
    assert "Tree" in result


@needs_314
def test_get_type_hint_uses_annotationlib_on_name_error() -> None:
    """_get_type_hint falls back to annotationlib.get_annotations on 3.14+ NameError."""
    sentinel = {"x": int}

    def dummy() -> None: ...

    with (
        patch(
            "sphinx_autodoc_typehints._resolver._type_hints.get_type_hints",
            side_effect=NameError("name 'Foo' is not defined"),
        ),
        patch.object(annotationlib, "get_annotations", return_value=sentinel) as mock_get_ann,
    ):
        result = _get_type_hint([], "dummy", dummy, {})

    mock_get_ann.assert_called_once_with(dummy, format=annotationlib.Format.FORWARDREF)
    assert result is sentinel


@pytest.mark.skipif(sys.version_info >= (3, 14), reason="Tests pre-3.14 fallback path")
def test_get_type_hint_falls_back_to_dunder_annotations_before_314() -> None:  # pragma: <3.14 cover
    """_get_type_hint falls back to __annotations__ on pre-3.14 NameError."""

    def dummy(x: int) -> str: ...

    with patch(
        "sphinx_autodoc_typehints._resolver._type_hints.get_type_hints",
        side_effect=NameError("name 'Foo' is not defined"),
    ):
        result = _get_type_hint([], "dummy", dummy, {})

    assert result == dummy.__annotations__



"""Module without deferred annotations to test the globals-scanning path."""

from typing import Any, Union

UserId = Union[int, str]
RequestData = dict[str, Any]


def get_user_eager(user_id: UserId) -> str:
    """
    Get a user by ID.

    Args:
        user_id: The user identifier
    """
    return f"User {user_id}"



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import MagicMock, create_autospec

from docutils import nodes
from sphinx.application import Sphinx
from sphinx.config import Config

import sphinx_autodoc_typehints as sat
from sphinx_autodoc_typehints._formats import InsertIndexInfo
from sphinx_autodoc_typehints._formats._sphinx import SphinxFieldListFormat, _node_line_no

if TYPE_CHECKING:
    import pytest


def test_inject_rtype_inserts_blank_line_before_rtype(monkeypatch: pytest.MonkeyPatch) -> None:
    def sample() -> str: ...

    config = create_autospec(
        Config,
        typehints_document_rtype=True,
        typehints_document_rtype_none=True,
        typehints_use_rtype=True,
        python_display_short_literal_types=False,
    )
    app: Sphinx = create_autospec(Sphinx, config=config)

    fmt = SphinxFieldListFormat()
    monkeypatch.setattr(
        fmt, "get_rtype_insert_info", lambda _app, _lines: InsertIndexInfo(insert_index=1, found_directive=True)
    )
    monkeypatch.setattr(sat, "format_annotation", lambda *_args, **_kwargs: "str")
    monkeypatch.setattr(sat, "add_type_css_class", lambda value: value)

    lines = ["A paragraph.", ".. note:: hi"]
    sat._inject_rtype({"return": str}, sample, app, "function", "sample", lines, fmt)  # noqa: SLF001

    assert lines == ["A paragraph.", "", ":rtype: str", "", ".. note:: hi"]


def test_inject_rtype_does_not_add_extra_blank_line(monkeypatch: pytest.MonkeyPatch) -> None:
    def sample() -> str: ...

    config = create_autospec(
        Config,
        typehints_document_rtype=True,
        typehints_document_rtype_none=True,
        typehints_use_rtype=True,
        python_display_short_literal_types=False,
    )
    app: Sphinx = create_autospec(Sphinx, config=config)

    fmt = SphinxFieldListFormat()
    monkeypatch.setattr(fmt, "get_rtype_insert_info", lambda _app, _lines: InsertIndexInfo(insert_index=1))
    monkeypatch.setattr(sat, "format_annotation", lambda *_args, **_kwargs: "str")
    monkeypatch.setattr(sat, "add_type_css_class", lambda value: value)

    lines = ["", ""]
    sat._inject_rtype({"return": str}, sample, app, "function", "sample", lines, fmt)  # noqa: SLF001

    assert lines == ["", ":rtype: str", ""]


def test_node_line_no_none_input() -> None:
    assert _node_line_no(None) is None  # type: ignore[arg-type]


def test_node_line_no_descends_to_find_line() -> None:
    child = MagicMock(spec=nodes.Node)
    child.line = 5
    child.children = []

    parent = MagicMock(spec=nodes.Node)
    parent.line = None
    parent.children = [child]

    assert _node_line_no(parent) == 5


def test_sphinx_format_detect_always_true() -> None:
    assert SphinxFieldListFormat.detect([]) is True
    assert SphinxFieldListFormat.detect([":param x: foo"]) is True


def test_get_rtype_insert_info_field_list_without_param_synonyms() -> None:
    """Line 169: field list exists but has no param-like fields, so it's skipped."""
    config = create_autospec(Config)
    app: Sphinx = create_autospec(Sphinx, config=config)
    app.env = MagicMock()

    fmt = SphinxFieldListFormat()
    lines = [":var x: some variable"]
    result = fmt.get_rtype_insert_info(app, lines)
    assert result is not None
    assert result.insert_index == len(lines)


def test_get_rtype_insert_info_directive_with_nonempty_preceding_line() -> None:
    """Line 182: directive found but preceding line is non-empty triggers break."""
    config = create_autospec(Config)
    app: Sphinx = create_autospec(Sphinx, config=config)
    app.env = MagicMock()

    fmt = SphinxFieldListFormat()
    lines = ["Some text.", "", ".. note::", "", "   Important note"]
    result = fmt.get_rtype_insert_info(app, lines)
    assert result is not None
    assert result.insert_index == len(lines)



"""Stub file (.pyi) annotation backfill."""

from __future__ import annotations

import ast
import inspect
from pathlib import Path
from typing import Any

from ._type_comments import _load_args

_STUB_AST_CACHE: dict[Path, ast.Module | None] = {}


def _backfill_from_stub(obj: Any) -> dict[str, str]:
    if (stub_path := _find_stub_path(obj)) and (tree := _parse_stub_ast(stub_path)):
        return _extract_annotations_from_stub(tree, obj)
    return {}


def _find_stub_path(obj: Any) -> Path | None:
    module = inspect.getmodule(obj)
    if module is None:
        return None
    try:
        source_file = inspect.getfile(module)
    except TypeError:
        return None
    stub = Path(source_file).with_suffix(".pyi")
    if stub.is_file():
        return stub
    if hasattr(module, "__path__"):
        for pkg_dir in module.__path__:
            init_stub = Path(pkg_dir) / "__init__.pyi"
            if init_stub.is_file():
                return init_stub
    return None


def _parse_stub_ast(stub_path: Path) -> ast.Module | None:
    if stub_path not in _STUB_AST_CACHE:
        try:
            _STUB_AST_CACHE[stub_path] = ast.parse(stub_path.read_text(encoding="utf-8"))
        except (OSError, SyntaxError):
            _STUB_AST_CACHE[stub_path] = None
    return _STUB_AST_CACHE[stub_path]


def _extract_annotations_from_stub(tree: ast.Module, obj: Any) -> dict[str, str]:
    qualname = getattr(obj, "__qualname__", None)
    if not qualname:
        return {}
    parts = qualname.split(".")
    if (node := _find_ast_node(tree.body, parts)) is None:
        return {}
    if isinstance(node, ast.ClassDef):
        return _extract_class_annotations(node)
    if isinstance(node, ast.FunctionDef | ast.AsyncFunctionDef):
        return _extract_func_annotations(node)
    return {}  # pragma: no cover


def _find_ast_node(body: list[ast.stmt], parts: list[str]) -> ast.stmt | None:
    target, *rest = parts
    for node in body:
        if isinstance(node, ast.ClassDef) and node.name == target:
            return _find_ast_node(node.body, rest) if rest else node
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)) and node.name == target and not rest:
            return node
    return None


def _extract_func_annotations(node: ast.FunctionDef | ast.AsyncFunctionDef) -> dict[str, str]:
    result: dict[str, str] = {}
    for arg in _load_args(node):
        if arg.annotation is not None:
            result[arg.arg] = ast.unparse(arg.annotation)
    if node.returns is not None:
        result["return"] = ast.unparse(node.returns)
    return result


def _extract_class_annotations(node: ast.ClassDef) -> dict[str, str]:
    result: dict[str, str] = {}
    for child in node.body:
        if isinstance(child, ast.AnnAssign) and isinstance(child.target, ast.Name):
            result[child.target.id] = ast.unparse(child.annotation)
    return result


__all__: list[str] = []



from __future__ import annotations

import enum
import re
import sys
import types
import typing
from collections.abc import Callable, Mapping
from io import StringIO
from types import EllipsisType, FrameType, FunctionType, ModuleType, NotImplementedType, TracebackType
from typing import (  # noqa: UP035
    IO,
    TYPE_CHECKING,
    Annotated,
    Any,
    AnyStr,
    Dict,
    ForwardRef,
    Generic,
    List,
    Literal,
    NewType,
    NotRequired,
    Optional,
    Required,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from unittest.mock import create_autospec

import pytest
import typing_extensions
from sphinx.config import Config

from sphinx_autodoc_typehints import (
    format_annotation,
    get_annotation_args,
    get_annotation_class_name,
    get_annotation_module,
)

if TYPE_CHECKING:
    from sphobjinv import Inventory

T = TypeVar("T")
U_co = TypeVar("U_co", covariant=True)
V_contra = TypeVar("V_contra", contravariant=True)
X = TypeVar("X", str, int)
Y = TypeVar("Y", bound=str)
Z = TypeVar("Z", bound="A")
S = TypeVar("S", bound="miss")  # type: ignore[name-defined] # miss not defined on purpose # noqa: F821
W = NewType("W", str)


class SomeEnum(enum.Enum):
    VALUE = "val"


P = typing_extensions.ParamSpec("P")
P_args = P.args
P_kwargs = P.kwargs
P_co = typing_extensions.ParamSpec("P_co", covariant=True)  # ty: ignore[invalid-paramspec]
P_contra = typing_extensions.ParamSpec("P_contra", contravariant=True)  # ty: ignore[invalid-paramspec]
P_bound = typing_extensions.ParamSpec("P_bound", bound=str)  # ty: ignore[invalid-paramspec]
RecList = Union[int, List["RecList"]]
MutualRecA = Union[bool, List["MutualRecB"]]
MutualRecB = Union[str, List["MutualRecA"]]


class A:
    def get_type(self) -> type:
        return type(self)  # pragma: no cover

    class Inner: ...


class B[T]:
    name = "Foo"


class C(B[str]): ...


class D(typing_extensions.Protocol): ...


class E(typing_extensions.Protocol[T]): ...


class Slotted:
    __slots__ = ()


class Metaclass(type): ...


PY312_PLUS = sys.version_info >= (3, 12)


@pytest.mark.parametrize(
    ("annotation", "module", "class_name", "args"),
    [
        pytest.param(str, "builtins", "str", (), id="str"),
        pytest.param(None, "builtins", "None", (), id="None"),
        pytest.param(ModuleType, "types", "ModuleType", (), id="ModuleType"),
        pytest.param(FunctionType, "types", "FunctionType", (), id="FunctionType"),
        pytest.param(types.CodeType, "types", "CodeType", (), id="CodeType"),
        pytest.param(types.CoroutineType, "types", "CoroutineType", (), id="CoroutineType"),
        pytest.param(Any, "typing", "Any", (), id="Any"),
        pytest.param(AnyStr, "typing", "AnyStr", (), id="AnyStr"),
        pytest.param(Dict, "typing", "Dict", (), id="Dict"),
        pytest.param(Dict[str, int], "typing", "Dict", (str, int), id="Dict_parametrized"),
        pytest.param(Dict[T, int], "typing", "Dict", (T, int), id="Dict_typevar"),
        pytest.param(Tuple, "typing", "Tuple", (), id="Tuple"),
        pytest.param(Tuple[str, int], "typing", "Tuple", (str, int), id="Tuple_parametrized"),
        pytest.param(Union[str, int], "typing", "Union", (str, int), id="Union"),
        pytest.param(Callable, "collections.abc", "Callable", (), id="Callable"),
        pytest.param(Callable[..., str], "collections.abc", "Callable", (..., str), id="Callable_returntype"),
        pytest.param(
            Callable[[int, str], str], "collections.abc", "Callable", (int, str, str), id="Callable_all_types"
        ),
        pytest.param(
            Callable[[int, str], str],
            "collections.abc",
            "Callable",
            (int, str, str),
            id="collections.abc.Callable_all_types",
        ),
        pytest.param(re.Pattern, "re", "Pattern", (), id="Pattern"),
        pytest.param(re.Pattern[str], "re", "Pattern", (str,), id="Pattern_parametrized"),
        pytest.param(re.Match, "re", "Match", (), id="Match"),
        pytest.param(re.Match[str], "re", "Match", (str,), id="Match_parametrized"),
        pytest.param(IO, "typing", "IO", (), id="IO"),
        pytest.param(W, "typing", "NewType", (str,), id="W"),
        pytest.param(P, "typing", "ParamSpec", (), id="P"),
        pytest.param(P_args, "typing", "ParamSpecArgs", (), id="P_args"),
        pytest.param(P_kwargs, "typing", "ParamSpecKwargs", (), id="P_kwargs"),
        pytest.param(Metaclass, __name__, "Metaclass", (), id="Metaclass"),
        pytest.param(Slotted, __name__, "Slotted", (), id="Slotted"),
        pytest.param(A, __name__, "A", (), id="A"),
        pytest.param(B, __name__, "B", (), id="B"),
        pytest.param(C, __name__, "C", (), id="C"),
        pytest.param(D, __name__, "D", (), id="D"),
        pytest.param(E, __name__, "E", (), id="E"),
        pytest.param(E[int], __name__, "E", (int,), id="E_parametrized"),
        pytest.param(A.Inner, __name__, "A.Inner", (), id="Inner"),
    ],
)
def test_parse_annotation(annotation: Any, module: str, class_name: str, args: tuple[Any, ...]) -> None:
    got_mod = get_annotation_module(annotation)
    got_cls = get_annotation_class_name(annotation, module)
    got_args = get_annotation_args(annotation, module, class_name)
    assert (got_mod, got_cls, got_args) == (module, class_name, args)


_CASES = [
    pytest.param(str, ":py:class:`str`", id="str"),
    pytest.param(int, ":py:class:`int`", id="int"),
    pytest.param(StringIO, ":py:class:`~io.StringIO`", id="StringIO"),
    pytest.param(EllipsisType, ":py:data:`~types.EllipsisType`", id="EllipsisType"),
    pytest.param(FunctionType, ":py:data:`~types.FunctionType`", id="FunctionType"),
    pytest.param(FrameType, ":py:data:`~types.FrameType`", id="FrameType"),
    pytest.param(ModuleType, ":py:class:`~types.ModuleType`", id="ModuleType"),
    pytest.param(NotImplementedType, ":py:data:`~types.NotImplementedType`", id="NotImplementedType"),
    pytest.param(TracebackType, ":py:class:`~types.TracebackType`", id="TracebackType"),
    pytest.param(type(None), ":py:obj:`None`", id="type None"),
    pytest.param(type, ":py:class:`type`", id="type"),
    pytest.param(Callable, ":py:class:`~collections.abc.Callable`", id="abc-Callable"),
    pytest.param(Type, ":py:class:`~typing.Type`", id="typing-Type"),
    pytest.param(Type[A], rf":py:class:`~typing.Type`\ \[:py:class:`~{__name__}.A`]", id="typing-A"),
    pytest.param(Any, ":py:data:`~typing.Any`", id="Any"),
    pytest.param(AnyStr, ":py:data:`~typing.AnyStr`", id="AnyStr"),
    pytest.param(Generic[T], r":py:class:`~typing.Generic`\ \[:py:class:`~typing.TypeVar`\ \(``T``)]", id="Generic"),
    pytest.param(Mapping, ":py:class:`~collections.abc.Mapping`", id="Mapping"),
    pytest.param(
        Mapping[T, int],
        r":py:class:`~collections.abc.Mapping`\ \[:py:class:`~typing.TypeVar`\ \(``T``), :py:class:`int`]",
        id="Mapping-T-int",
    ),
    pytest.param(
        Mapping[str, V_contra],
        r":py:class:`~collections.abc.Mapping`\ \[:py:class:`str`, :py:class:`~typing.TypeVar`\ \("
        "``V_contra``, contravariant=True)]",
        id="Mapping-T-int-contra",
    ),
    pytest.param(
        Mapping[T, U_co],
        r":py:class:`~collections.abc.Mapping`\ \[:py:class:`~typing.TypeVar`\ \(``T``), "
        r":py:class:`~typing.TypeVar`\ \(``U_co``, covariant=True)]",
        id="Mapping-T-int-co",
    ),
    pytest.param(
        Mapping[str, bool],
        r":py:class:`~collections.abc.Mapping`\ \[:py:class:`str`, :py:class:`bool`]",
        id="Mapping-str-bool",
    ),
    pytest.param(Dict, ":py:class:`~typing.Dict`", id="Dict"),
    pytest.param(
        Dict[T, int],
        r":py:class:`~typing.Dict`\ \[:py:class:`~typing.TypeVar`\ \(``T``), :py:class:`int`]",
        id="Dict-T-int",
    ),
    pytest.param(
        Dict[str, V_contra],
        r":py:class:`~typing.Dict`\ \[:py:class:`str`, :py:class:`~typing.TypeVar`\ \(``V_contra``, "
        r"contravariant=True)]",
        id="Dict-T-int-contra",
    ),
    pytest.param(
        Dict[T, U_co],
        r":py:class:`~typing.Dict`\ \[:py:class:`~typing.TypeVar`\ \(``T``),"
        r" :py:class:`~typing.TypeVar`\ \(``U_co``, covariant=True)]",
        id="Dict-T-int-co",
    ),
    pytest.param(
        Dict[str, bool],
        r":py:class:`~typing.Dict`\ \[:py:class:`str`, :py:class:`bool`]",
        id="Dict-str-bool",
    ),
    pytest.param(Tuple, ":py:data:`~typing.Tuple`", id="Tuple"),
    pytest.param(
        Tuple[str, bool],
        r":py:data:`~typing.Tuple`\ \[:py:class:`str`, :py:class:`bool`]",
        id="Tuple-str-bool",
    ),
    pytest.param(
        Tuple[int, int, int],
        r":py:data:`~typing.Tuple`\ \[:py:class:`int`, :py:class:`int`, :py:class:`int`]",
        id="Tuple-int-int-int",
    ),
    pytest.param(
        Tuple[str, ...],
        r":py:data:`~typing.Tuple`\ \[:py:class:`str`, :py:data:`...<Ellipsis>`]",
        id="Tuple-str-Ellipsis",
    ),
    pytest.param(Union, f":py:{'class' if sys.version_info >= (3, 14) else 'data'}:`~typing.Union`", id="Union"),
    pytest.param(
        types.UnionType, f":py:{'class' if sys.version_info >= (3, 14) else 'data'}:`~typing.Union`", id="UnionType"
    ),
    pytest.param(
        Union[str, bool],
        ":py:class:`str` | :py:class:`bool`"
        if sys.version_info >= (3, 14)
        else r":py:data:`~typing.Union`\ \[:py:class:`str`, :py:class:`bool`]",
        id="Union-str-bool",
    ),
    pytest.param(
        Union[str, bool, None],
        ":py:class:`str` | :py:class:`bool` | :py:obj:`None`"
        if sys.version_info >= (3, 14)
        else r":py:data:`~typing.Union`\ \[:py:class:`str`, :py:class:`bool`, :py:obj:`None`]",
        id="Union-str-bool-None",
    ),
    pytest.param(
        Union[str, Any],
        ":py:class:`str` | :py:data:`~typing.Any`"
        if sys.version_info >= (3, 14)
        else r":py:data:`~typing.Union`\ \[:py:class:`str`, :py:data:`~typing.Any`]",
        id="Union-str-Any",
    ),
    pytest.param(
        Optional[str],
        ":py:class:`str` | :py:obj:`None`"
        if sys.version_info >= (3, 14)
        else r":py:data:`~typing.Optional`\ \[:py:class:`str`]",
        id="Optional-str",
    ),
    pytest.param(
        Union[str, None],
        ":py:class:`str` | :py:obj:`None`"
        if sys.version_info >= (3, 14)
        else r":py:data:`~typing.Optional`\ \[:py:class:`str`]",
        id="Optional-str-None",
    ),
    pytest.param(
        type[T] | types.UnionType,
        ":py:class:`type`\\ \\[:py:class:`~typing.TypeVar`\\ \\(``T``)] | "
        f":py:{'class' if sys.version_info >= (3, 14) else 'data'}:`~typing.Union`",
        id="typevar union bar uniontype",
    ),
    pytest.param(
        Optional[str | bool],
        ":py:class:`str` | :py:class:`bool` | :py:obj:`None`"
        if sys.version_info >= (3, 14)
        else r":py:data:`~typing.Union`\ \[:py:class:`str`, :py:class:`bool`, :py:obj:`None`]",
        id="Optional-Union-str-bool",
    ),
    pytest.param(
        RecList,
        ":py:class:`int` | :py:class:`~typing.List`\\ \\[RecList]"
        if sys.version_info >= (3, 14)
        else r":py:data:`~typing.Union`\ \[:py:class:`int`, :py:class:`~typing.List`\ \[RecList]]",
        id="RecList",
    ),
    pytest.param(
        MutualRecA,
        ":py:class:`bool` | :py:class:`~typing.List`\\ \\[MutualRecB]"
        if sys.version_info >= (3, 14)
        else r":py:data:`~typing.Union`\ \[:py:class:`bool`, :py:class:`~typing.List`\ \[MutualRecB]]",
        id="MutualRecA",
    ),
    pytest.param(Callable, ":py:class:`~collections.abc.Callable`", id="Callable"),
    pytest.param(
        Callable[..., int],
        r":py:class:`~collections.abc.Callable`\ \[:py:data:`...<Ellipsis>`, :py:class:`int`]",
        id="Callable-Ellipsis-int",
    ),
    pytest.param(
        Callable[[int], int],
        r":py:class:`~collections.abc.Callable`\ \[\[:py:class:`int`], :py:class:`int`]",
        id="Callable-int-int",
    ),
    pytest.param(
        Callable[[int, str], bool],
        r":py:class:`~collections.abc.Callable`\ \[\[:py:class:`int`, :py:class:`str`], :py:class:`bool`]",
        id="Callable-int-str-bool",
    ),
    pytest.param(
        Callable[[int, str], None],
        r":py:class:`~collections.abc.Callable`\ \[\[:py:class:`int`, :py:class:`str`], :py:obj:`None`]",
        id="Callable-int-str",
    ),
    pytest.param(
        Callable[[T], T],
        r":py:class:`~collections.abc.Callable`\ \[\[:py:class:`~typing.TypeVar`\ \(``T``)],"
        r" :py:class:`~typing.TypeVar`\ \(``T``)]",
        id="Callable-T-T",
    ),
    pytest.param(
        Callable[[int, str], bool],
        r":py:class:`~collections.abc.Callable`\ \[\[:py:class:`int`, :py:class:`str`], :py:class:`bool`]",
        id="Callable-int-str-bool",
    ),
    pytest.param(re.Pattern, ":py:class:`~re.Pattern`", id="Pattern"),
    pytest.param(re.Pattern[str], r":py:class:`~re.Pattern`\ \[:py:class:`str`]", id="Pattern-str"),
    pytest.param(IO, ":py:class:`~typing.IO`", id="IO"),
    pytest.param(IO[str], r":py:class:`~typing.IO`\ \[:py:class:`str`]", id="IO-str"),
    pytest.param(Metaclass, f":py:class:`~{__name__}.Metaclass`", id="Metaclass"),
    pytest.param(A, f":py:class:`~{__name__}.A`", id="A"),
    pytest.param(B, f":py:class:`~{__name__}.B`", id="B"),
    pytest.param(B[int], rf":py:class:`~{__name__}.B`\ \[:py:class:`int`]", id="B-int"),
    pytest.param(C, f":py:class:`~{__name__}.C`", id="C"),
    pytest.param(D, f":py:class:`~{__name__}.D`", id="D"),
    pytest.param(E, f":py:class:`~{__name__}.E`", id="E"),
    pytest.param(E[int], rf":py:class:`~{__name__}.E`\ \[:py:class:`int`]", id="E-int"),
    pytest.param(W, f":py:class:`~{__name__}.W` (:py:class:`str`)", id="W"),
    pytest.param(T, r":py:class:`~typing.TypeVar`\ \(``T``)", id="T"),
    pytest.param(U_co, r":py:class:`~typing.TypeVar`\ \(``U_co``, covariant=True)", id="U-co"),
    pytest.param(V_contra, r":py:class:`~typing.TypeVar`\ \(``V_contra``, contravariant=True)", id="V-contra"),
    pytest.param(X, r":py:class:`~typing.TypeVar`\ \(``X``, :py:class:`str`, :py:class:`int`)", id="X"),
    pytest.param(Y, r":py:class:`~typing.TypeVar`\ \(``Y``, bound= :py:class:`str`)", id="Y"),
    pytest.param(Z, r":py:class:`~typing.TypeVar`\ \(``Z``, bound= A)", id="Z"),
    pytest.param(S, r":py:class:`~typing.TypeVar`\ \(``S``, bound= miss)", id="S"),
    pytest.param(
        P, rf":py:class:`~typing.ParamSpec`\ \(``P``{', bound= :py:obj:`None`' if PY312_PLUS else ''})", id="P"
    ),
    pytest.param(
        P_co,
        rf":py:class:`~typing.ParamSpec`\ \(``P_co``{', bound= :py:obj:`None`' if PY312_PLUS else ''}, covariant=True)",
        id="P_co",
    ),
    pytest.param(
        P_contra,
        rf":py:class:`~typing.ParamSpec`\ \(``P_contra``{', bound= :py:obj:`None`' if PY312_PLUS else ''}"
        ", contravariant=True)",
        id="P-contra",
    ),
    pytest.param(P_bound, r":py:class:`~typing.ParamSpec`\ \(``P_bound``, bound= :py:class:`str`)", id="P-bound"),
    pytest.param(Tuple[()], ":py:data:`~typing.Tuple`", id="Tuple-p"),
    pytest.param(Tuple[int,], r":py:data:`~typing.Tuple`\ \[:py:class:`int`]", id="Tuple-p-int"),
    pytest.param(
        Tuple[int, int],
        r":py:data:`~typing.Tuple`\ \[:py:class:`int`, :py:class:`int`]",
        id="Tuple-p-int-int",
    ),
    pytest.param(
        Tuple[int, ...],
        r":py:data:`~typing.Tuple`\ \[:py:class:`int`, :py:data:`...<Ellipsis>`]",
        id="Tuple-p-Ellipsis",
    ),
    pytest.param(Annotated[int, "metadata"], r":py:class:`int`", id="Annotated-metadata"),
    pytest.param(Required[int], r":py:class:`~typing.Required`\ \[:py:class:`int`]", id="Required"),
    pytest.param(NotRequired[int], r":py:class:`~typing.NotRequired`\ \[:py:class:`int`]", id="NotRequired"),
]


@pytest.mark.parametrize(("annotation", "expected_result"), _CASES)
def test_format_annotation(inv: Inventory, annotation: Any, expected_result: str) -> None:
    conf = create_autospec(Config, _annotation_globals=globals(), always_use_bars_union=False)
    result = format_annotation(annotation, conf)
    assert result == expected_result

    if re.match(r"^:py:data:`~typing\.Union`\\\[.*``None``.*]", expected_result):  # pragma: <3.14 cover
        expected_result_not_simplified = expected_result.replace(", ``None``", "")
        expected_result_not_simplified += ":py:data:`~typing.Optional`\\ \\["
        expected_result_not_simplified += "]"
        conf = create_autospec(
            Config,
            simplify_optional_unions=False,
            _annotation_globals=globals(),
            always_use_bars_union=False,
        )
        assert format_annotation(annotation, conf) == expected_result_not_simplified

        if "typing" in expected_result_not_simplified:
            expected_result_not_simplified = expected_result_not_simplified.replace("~typing", "typing")
            conf = create_autospec(
                Config,
                typehints_fully_qualified=True,
                simplify_optional_unions=False,
                _annotation_globals=globals(),
            )
            assert format_annotation(annotation, conf) == expected_result_not_simplified

    if "typing" in expected_result or __name__ in expected_result:
        expected_result = expected_result.replace("~typing", "typing")
        expected_result = expected_result.replace("~collections.abc", "collections.abc")
        expected_result = expected_result.replace("~numpy", "numpy")
        expected_result = expected_result.replace("~" + __name__, __name__)
        conf = create_autospec(
            Config,
            typehints_fully_qualified=True,
            _annotation_globals=globals(),
            always_use_bars_union=False,
        )
        assert format_annotation(annotation, conf) == expected_result

    if (
        result.count(":py:") == 1
        and ("typing" in result or "types" in result)
        and (match := re.match(r"^:py:(?P<role>class|data|func):`~(?P<name>[^`]+)`", result))
    ):
        name = match.group("name")
        expected_role = next((o.role for o in inv.objects if o.name == name), None)
        if expected_role and expected_role == "function":  # pragma: no cover
            expected_role = "func"
        assert match.group("role") == expected_role


@pytest.mark.parametrize(
    ("annotation", "expected_result"),
    [
        ("int | float", ":py:class:`int` | :py:class:`float`"),
        ("int | float | None", ":py:class:`int` | :py:class:`float` | :py:obj:`None`"),
        ("Union[int, float]", ":py:class:`int` | :py:class:`float`"),
        ("Union[int, float, None]", ":py:class:`int` | :py:class:`float` | :py:obj:`None`"),
        ("Optional[int | float]", ":py:class:`int` | :py:class:`float` | :py:obj:`None`"),
        ("Optional[Union[int, float]]", ":py:class:`int` | :py:class:`float` | :py:obj:`None`"),
        ("Union[int | float, str]", ":py:class:`int` | :py:class:`float` | :py:class:`str`"),
        ("Union[int, float] | str", ":py:class:`int` | :py:class:`float` | :py:class:`str`"),
    ],
)
def test_always_use_bars_union(annotation: str, expected_result: str) -> None:
    conf = create_autospec(Config, always_use_bars_union=True)
    result = format_annotation(eval(annotation), conf)  # noqa: S307
    assert result == expected_result


@pytest.mark.parametrize("library", [typing, typing_extensions], ids=["typing", "typing_extensions"])
@pytest.mark.parametrize(
    ("annotation", "params", "expected_result"),
    [
        pytest.param("ClassVar", int, ":py:data:`~typing.ClassVar`\\ \\[:py:class:`int`]", id="ClassVar"),
        pytest.param("NoReturn", None, ":py:data:`~typing.NoReturn`", id="NoReturn"),
        pytest.param("Literal", ("a", 1), ":py:data:`~typing.Literal`\\ \\[``'a'``, ``1``]", id="Literal"),
        pytest.param(
            "Literal",
            (SomeEnum.VALUE,),
            rf":py:data:`~typing.Literal`\ \[:py:attr:`~{__name__}.SomeEnum.VALUE`]",
            id="Literal-enum",
        ),
        pytest.param("Type", None, ":py:class:`~typing.Type`", id="Type-none"),
        pytest.param("Type", (A,), rf":py:class:`~typing.Type`\ \[:py:class:`~{__name__}.A`]", id="Type-A"),
    ],
)
def test_format_annotation_both_libs(library: ModuleType, annotation: str, params: Any, expected_result: str) -> None:
    try:
        annotation_cls = getattr(library, annotation)
    except AttributeError:  # pragma: no cover
        pytest.skip(f"{annotation} not available in the {library.__name__} module")

    ann = annotation_cls if params is None else annotation_cls[params]
    result = format_annotation(ann, create_autospec(Config))
    assert result == expected_result


def test_format_annotation_tuple() -> None:
    conf = create_autospec(Config)
    assert format_annotation((int, str), conf) == "(:py:class:`int`, :py:class:`str`)"


def test_format_annotation_empty_tuple() -> None:
    conf = create_autospec(Config)
    assert format_annotation((), conf) == "()"


def test_format_annotation_single_element_tuple() -> None:
    conf = create_autospec(Config)
    assert format_annotation((int,), conf) == "(:py:class:`int`, )"


def test_format_annotation_none() -> None:
    conf = create_autospec(Config)
    assert format_annotation(None, conf) == ":py:obj:`None`"


def test_format_annotation_ellipsis() -> None:
    conf = create_autospec(Config)
    assert format_annotation(Ellipsis, conf) == ":py:data:`...<Ellipsis>`"


def test_format_annotation_forward_ref() -> None:
    conf = create_autospec(Config)
    assert format_annotation(ForwardRef("SomeClass"), conf) == "SomeClass"


def test_format_annotation_typing_extensions_module_fixup() -> None:
    conf = create_autospec(Config)
    result = format_annotation(typing_extensions.ClassVar[int], conf)
    assert "typing.ClassVar" in result


def test_format_annotation_io_module_fixup() -> None:
    conf = create_autospec(Config)
    result = format_annotation(StringIO, conf)
    assert "io.StringIO" in result


def test_format_annotation_with_formatter_returning_value() -> None:
    conf = create_autospec(Config, typehints_formatter=lambda ann, _cfg: f"Custom({ann})")
    result = format_annotation(int, conf)
    assert result == "Custom(<class 'int'>)"


def test_format_annotation_with_formatter_returning_none() -> None:
    conf = create_autospec(Config, typehints_formatter=lambda _ann, _cfg: None, always_use_bars_union=False)
    result = format_annotation(int, conf)
    assert result == ":py:class:`int`"


def test_format_annotation_short_literals() -> None:
    conf = create_autospec(Config)
    result = format_annotation(Literal["a", "b"], conf, short_literals=True)
    assert result == "\\``'a'`` | ``'b'``"


def test_get_annotation_module_raises_on_unknown() -> None:
    with pytest.raises(ValueError, match="Cannot determine the module"):
        get_annotation_module(42)


def test_get_annotation_class_name_name_attr() -> None:
    result = get_annotation_class_name(IO, "typing")
    assert result == "IO"


def test_get_annotation_args_classvar() -> None:
    result = get_annotation_args(typing.ClassVar[int], "typing", "ClassVar")
    assert result == (int,)


def test_get_annotation_args_literal_values() -> None:
    result = get_annotation_args(typing.Literal["a", 1], "typing", "Literal")
    assert result == ("a", 1)


def test_get_annotation_args_generic() -> None:
    result = get_annotation_args(Generic[T], "typing", "Generic")
    assert result == (T,)



from __future__ import annotations

import ast
import sys
import types
from pathlib import Path
from typing import TYPE_CHECKING, Any
from unittest.mock import MagicMock, patch

import pytest
from conftest import normalize_sphinx_text

from sphinx_autodoc_typehints._resolver._stubs import (
    _STUB_AST_CACHE,
    _backfill_from_stub,
    _extract_annotations_from_stub,
    _extract_class_annotations,
    _extract_func_annotations,
    _find_ast_node,
    _find_stub_path,
    _parse_stub_ast,
)

if TYPE_CHECKING:
    from io import StringIO

    from sphinx.testing.util import SphinxTestApp

STUB_ROOT = Path(__file__).parent.parent / "roots" / "test-pyi-stubs"


def _import_stub_mod() -> types.ModuleType:
    import stub_mod  # noqa: PLC0415  # ty: ignore[unresolved-import]

    return stub_mod


@pytest.fixture
def stub_mod() -> Any:
    sys.path.insert(0, str(STUB_ROOT))
    try:
        yield _import_stub_mod()
    finally:
        sys.path.pop(0)
        sys.modules.pop("stub_mod", None)


def test_find_stub_path_locates_sibling_pyi(stub_mod: Any) -> None:
    result = _find_stub_path(stub_mod.greet)
    assert result is not None
    assert result.suffix == ".pyi"
    assert result.stem == "stub_mod"


def test_find_stub_path_returns_none_for_no_stub() -> None:
    assert _find_stub_path(test_find_stub_path_returns_none_for_no_stub) is None


def test_find_stub_path_returns_none_for_no_module() -> None:
    obj = MagicMock(spec=[])
    obj.__module__ = None
    with patch("sphinx_autodoc_typehints._resolver._stubs.inspect.getmodule", return_value=None):
        assert _find_stub_path(obj) is None


def test_find_stub_path_returns_none_when_getfile_fails() -> None:
    module = types.ModuleType("fake_mod")
    with (
        patch("sphinx_autodoc_typehints._resolver._stubs.inspect.getmodule", return_value=module),
        patch("sphinx_autodoc_typehints._resolver._stubs.inspect.getfile", side_effect=TypeError),
    ):
        assert _find_stub_path(lambda: None) is None


def test_find_stub_path_package_init_pyi(tmp_path: Path) -> None:
    pkg_dir = tmp_path / "mypkg"
    pkg_dir.mkdir()
    (pkg_dir / "__init__.py").write_text("")
    no_stub_dir = tmp_path / "no_stubs"
    no_stub_dir.mkdir()
    stub_dir = tmp_path / "stubs"
    stub_dir.mkdir()
    (stub_dir / "__init__.pyi").write_text("x: int\n")
    module = types.ModuleType("mypkg")
    module.__path__ = [str(no_stub_dir), str(stub_dir)]
    module.__file__ = str(pkg_dir / "__init__.py")
    with patch("sphinx_autodoc_typehints._resolver._stubs.inspect.getmodule", return_value=module):
        result = _find_stub_path(module)
    assert result is not None
    assert result.name == "__init__.pyi"


def test_find_stub_path_package_no_init_pyi(tmp_path: Path) -> None:
    pkg_dir = tmp_path / "mypkg"
    pkg_dir.mkdir()
    (pkg_dir / "__init__.py").write_text("")
    module = types.ModuleType("mypkg")
    module.__path__ = [str(pkg_dir)]
    module.__file__ = str(pkg_dir / "__init__.py")
    with patch("sphinx_autodoc_typehints._resolver._stubs.inspect.getmodule", return_value=module):
        assert _find_stub_path(module) is None


def test_parse_stub_ast_valid_file() -> None:
    stub_path = STUB_ROOT / "stub_mod.pyi"
    result = _parse_stub_ast(stub_path)
    assert isinstance(result, ast.Module)


def test_parse_stub_ast_caches_result() -> None:
    stub_path = STUB_ROOT / "stub_mod.pyi"
    _STUB_AST_CACHE.pop(stub_path, None)
    first = _parse_stub_ast(stub_path)
    second = _parse_stub_ast(stub_path)
    assert first is second
    _STUB_AST_CACHE.pop(stub_path, None)


def test_parse_stub_ast_caches_none_for_bad_syntax(tmp_path: Path) -> None:
    bad_stub = tmp_path / "bad.pyi"
    bad_stub.write_text("def (broken syntax\n")
    result = _parse_stub_ast(bad_stub)
    assert result is None
    assert bad_stub in _STUB_AST_CACHE
    assert _STUB_AST_CACHE[bad_stub] is None
    _STUB_AST_CACHE.pop(bad_stub, None)


def test_parse_stub_ast_returns_none_for_missing_file() -> None:
    missing = Path("/nonexistent/stub.pyi")
    _STUB_AST_CACHE.pop(missing, None)
    assert _parse_stub_ast(missing) is None
    _STUB_AST_CACHE.pop(missing, None)


@pytest.mark.parametrize(
    ("source", "parts", "expected_type", "expected_name"),
    [
        pytest.param("def foo(x: int) -> str: ...", ["foo"], ast.FunctionDef, "foo", id="top_level_function"),
        pytest.param(
            "class Outer:\n  class Inner:\n    def method(self) -> None: ...",
            ["Outer", "Inner", "method"],
            ast.FunctionDef,
            "method",
            id="nested_class_method",
        ),
        pytest.param("class Foo:\n  x: int\n", ["Foo"], ast.ClassDef, "Foo", id="class"),
        pytest.param("def foo(): ...", ["bar"], None, None, id="missing"),
    ],
)
def test_find_ast_node(
    source: str, parts: list[str], expected_type: type[ast.stmt] | None, expected_name: str | None
) -> None:
    tree = ast.parse(source)
    node = _find_ast_node(tree.body, parts)
    if expected_type is None:
        assert node is None
    else:
        assert isinstance(node, expected_type)
        assert node.name == expected_name  # type: ignore[union-attr]


@pytest.mark.parametrize(
    ("source", "expected"),
    [
        pytest.param(
            "def greet(name: str, greeting: str) -> str: ...",
            {"name": "str", "greeting": "str", "return": "str"},
            id="basic",
        ),
        pytest.param("def f(x: int): ...", {"x": "int"}, id="no_return"),
        pytest.param("async def fetch(url: str) -> str: ...", {"url": "str", "return": "str"}, id="async"),
        pytest.param("def f(self, x: int) -> None: ...", {"x": "int", "return": "None"}, id="skips_unannotated"),
    ],
)
def test_extract_func_annotations(source: str, expected: dict[str, str]) -> None:
    node = ast.parse(source).body[0]
    assert isinstance(node, ast.FunctionDef | ast.AsyncFunctionDef)
    assert _extract_func_annotations(node) == expected


@pytest.mark.parametrize(
    ("source", "expected"),
    [
        pytest.param("class Foo:\n  x: int\n  y: str\n", {"x": "int", "y": "str"}, id="basic"),
        pytest.param("class Foo:\n  x: int\n  self.y: str\n", {"x": "int"}, id="ignores_non_name_targets"),
    ],
)
def test_extract_class_annotations(source: str, expected: dict[str, str]) -> None:
    node = ast.parse(source).body[0]
    assert isinstance(node, ast.ClassDef)
    assert _extract_class_annotations(node) == expected


@pytest.mark.parametrize(
    ("source", "qualname", "expected"),
    [
        pytest.param(
            "def greet(name: str, greeting: str) -> str: ...",
            "greet",
            {"name": "str", "greeting": "str", "return": "str"},
            id="function",
        ),
        pytest.param("class Foo:\n  x: int\n  y: str\n", "Foo", {"x": "int", "y": "str"}, id="class"),
        pytest.param("def greet(name: str) -> str: ...", "nonexistent", {}, id="missing_node"),
        pytest.param("x: int = 1\n", "x", {}, id="unsupported_node_type"),
    ],
)
def test_extract_annotations_from_stub(source: str, qualname: str, expected: dict[str, str]) -> None:
    tree = ast.parse(source)
    obj = MagicMock()
    obj.__qualname__ = qualname
    assert _extract_annotations_from_stub(tree, obj) == expected


def test_extract_annotations_from_stub_no_qualname() -> None:
    tree = ast.parse("def greet(name: str) -> str: ...")
    obj = MagicMock(spec=[])
    assert _extract_annotations_from_stub(tree, obj) == {}


@pytest.mark.parametrize(
    ("attr", "expected"),
    [
        pytest.param("greet", {"name": "str", "greeting": "str", "return": "str"}, id="function"),
        pytest.param("Calculator.Inner.process", {"data": "bytes", "return": "bytes"}, id="nested_class"),
        pytest.param("fetch", {"url": "str", "return": "str"}, id="async_function"),
    ],
)
def test_backfill_from_stub(stub_mod: Any, attr: str, expected: dict[str, str]) -> None:
    obj = stub_mod
    for part in attr.split("."):
        obj = getattr(obj, part)
    assert _backfill_from_stub(obj) == expected


def test_backfill_from_stub_no_stub() -> None:
    assert _backfill_from_stub(test_backfill_from_stub_no_stub) == {}


@pytest.mark.sphinx("text", testroot="pyi-stubs")
def test_sphinx_build_uses_stub_types(app: SphinxTestApp, status: StringIO, warning: StringIO) -> None:
    template = """\
.. autofunction:: stub_mod.greet

.. autoclass:: stub_mod.Calculator
   :members:

.. autofunction:: stub_mod.fetch
"""
    (Path(app.srcdir) / "index.rst").write_text(template)
    app.build()
    assert "build succeeded" in status.getvalue()
    result = normalize_sphinx_text((Path(app.srcdir) / "_build/text/index.txt").read_text())
    assert "str" in result
    warn_text = warning.getvalue()
    assert "stub_mod" not in warn_text or "forward reference" not in warn_text



from __future__ import annotations

import csv
from csv import Error
from typing import Any
from unittest.mock import MagicMock, patch

from sphinx_autodoc_typehints._resolver._type_hints import (
    _execute_guarded_code,
    _future_annotations_imported,
    _get_type_hint,
    _resolve_type_guarded_imports,
    _run_guarded_import,
    _should_skip_guarded_import_resolution,
)


def test_no_source_code_type_guard() -> None:
    _resolve_type_guarded_imports([], Error)


def test_future_annotations_not_imported() -> None:
    assert not _future_annotations_imported(csv)


def test_future_annotations_imported() -> None:
    assert _future_annotations_imported(test_future_annotations_imported)


def test_should_skip_module_type() -> None:
    assert not _should_skip_guarded_import_resolution(csv)


def test_should_skip_no_globals() -> None:
    assert _should_skip_guarded_import_resolution(42)


def test_should_skip_builtin_module() -> None:
    fn: Any = type("FakeFunc", (), {"__globals__": {}, "__module__": "builtins"})()
    assert _should_skip_guarded_import_resolution(fn)


def test_get_type_hint_recursion_error() -> None:
    def func(x: int) -> str: ...

    with patch("sphinx_autodoc_typehints._resolver._type_hints.get_type_hints", side_effect=RecursionError):
        assert _get_type_hint([], "test", func, {}) == {}


def test_execute_guarded_code_catches_exception() -> None:
    module = type("FakeModule", (), {"__globals__": {}, "__dict__": {}})()
    with patch("sphinx_autodoc_typehints._resolver._type_hints._run_guarded_import", side_effect=RuntimeError("boom")):
        _execute_guarded_code([], module, "\nif TYPE_CHECKING:\n    import os\nx = 1\n")


def test_run_guarded_import_no_exc_name() -> None:
    ns: dict[str, Any] = {}
    obj: Any = type("FakeObj", (), {"__globals__": ns})()
    _run_guarded_import([], obj, "raise ImportError()")


def test_forward_ref_warning_includes_module() -> None:
    def func(x: int) -> str: ...

    func.__module__ = "some_module"
    func.__annotations__ = {"x": "NonExistent"}
    mock_logger = MagicMock()
    with (
        patch("sphinx_autodoc_typehints._resolver._type_hints.get_type_hints", side_effect=NameError("NonExistent")),
        patch("sphinx_autodoc_typehints._resolver._type_hints._LOGGER", mock_logger),
    ):
        _get_type_hint([], "func", func, {})
    mock_logger.warning.assert_called_once()
    args = mock_logger.warning.call_args
    assert "some_module" in str(args)
    assert "location" in args.kwargs


def test_guarded_import_warning_includes_module() -> None:
    module = type("FakeModule", (), {"__globals__": {}, "__dict__": {}, "__module__": "fake_mod"})()
    mock_logger = MagicMock()
    with (
        patch("sphinx_autodoc_typehints._resolver._type_hints._run_guarded_import", side_effect=RuntimeError("boom")),
        patch("sphinx_autodoc_typehints._resolver._type_hints._LOGGER", mock_logger),
    ):
        _execute_guarded_code([], module, "\nif TYPE_CHECKING:\n    import os\nx = 1\n")
    mock_logger.warning.assert_called_once()
    args = mock_logger.warning.call_args
    assert "fake_mod" in str(args)



"""Tests that snippet parsing doesn't trigger extension directive side-effects."""

from __future__ import annotations

import sys
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING, ClassVar

import pytest
from docutils.parsers.rst import Directive, directives

if TYPE_CHECKING:
    from io import StringIO

    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx("text", testroot="integration")
def test_extension_directive_not_executed_during_snippet_parse(
    app: SphinxTestApp,
    status: StringIO,
    warning: StringIO,  # noqa: ARG001
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """A non-builtin directive in a docstring should only execute once (during the real build)."""
    directives.register_directive("tracking-directive", _TrackingDirective)
    _TrackingDirective.executions.clear()

    (Path(app.srcdir) / "index.rst").write_text(
        dedent("""\
        Test
        ====

        .. autofunction:: mod.func_with_tracking_directive
    """)
    )

    src = dedent("""\
    def func_with_tracking_directive(x: int) -> int:
        \"\"\"Do something.

        :param x: A number.

        .. tracking-directive::

            unique-id-123

        \"\"\"
        return x
    """)
    exec(compile(src, "<test>", "exec"), (mod := {}))  # noqa: S102
    fake_module = type(sys)("mod")
    fake_module.__dict__.update(mod)
    monkeypatch.setitem(sys.modules, "mod", fake_module)

    app.build()
    assert "build succeeded" in status.getvalue()
    assert _TrackingDirective.executions.count("unique-id-123") == 1


class _TrackingDirective(Directive):
    """Directive that records each execution to detect double-processing."""

    has_content = True
    executions: ClassVar[list[str]] = []

    def run(self) -> list:
        _TrackingDirective.executions.append(self.content[0] if self.content else "")
        return []



from __future__ import annotations

import attr
import attrs


@attr.s
class ClassicAttrs:
    """A class using classic attr.s style."""

    name = attr.ib(type=str)
    age = attr.ib(type=int)
    untyped = attr.ib()


@attr.s(auto_attribs=True)
class AutoAttribs:
    """A class using auto_attribs=True."""

    name: str
    age: int


@attrs.define
class ModernAttrs:
    """A class using modern attrs.define."""

    name: str
    age: int
